import numpy as np
import sys
import os
import csv
import tempfile 
import copy 
import pandas as pd
import sys, os
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

from basinmaker.func.pdtable import *
from osgeo import gdal, ogr

def decode_hru_attri_ids(HRU_temp1,lakehruinfo,Landuse_ID,Soil_ID,Veg_ID,Other_Ply_ID_1,Other_Ply_ID_2):
    #     100000000000000, 
    #          0007422000,
    #          2147483647,
#    'sub':       1000000,
#    'landuse':     10000,
#    'soil':          100,
#    'o1':              1,

    lakehruinfo_id = lakehruinfo[['HRULake_ID','SubId','HRU_IsLake']]
    # remove polygon outside of the subbasin 
    HRU_temp1 = HRU_temp1[HRU_temp1['HRU_ID_T'] != -9999]
    HRU_temp1['HRU_ID_str'] = HRU_temp1['HRU_ID_T'].astype('int64').astype(str)
#    HRU_temp1['HRU_ID_str'] = HRU_temp1['HRU_ID_T'].str.zfill(10)
           
    HRU_temp1['HRULake_ID'] = HRU_temp1['HRU_ID_str'].str[:-6].astype(int)
    HRU_temp1 = pd.merge(HRU_temp1, lakehruinfo_id, how='inner', on = 'HRULake_ID')
    HRU_temp1[Landuse_ID] = HRU_temp1['HRU_ID_str'].str[-6:-4].astype(int)
    HRU_temp1[Veg_ID] = HRU_temp1['HRU_ID_str'].str[-6:-4].astype(int)
    HRU_temp1[Soil_ID] = HRU_temp1['HRU_ID_str'].str[-4:-2].astype(int)
    HRU_temp1[Other_Ply_ID_1] = HRU_temp1['HRU_ID_str'].str[-2:].astype(int)
    HRU_temp1[Other_Ply_ID_2] = HRU_temp1['HRU_ID_str'].str[-2:].astype(int)
    HRU_temp1 = HRU_temp1.loc[(HRU_temp1['HRULake_ID'] > 0) | (HRU_temp1['SubId'] > 0)]
    
    return HRU_temp1
       
def raster_to_vector(output,input,raster_par_list):
    x_min = raster_par_list[0]
    x_max = raster_par_list[1]
    y_min = raster_par_list[2]
    y_max = raster_par_list[3]
    x_res = raster_par_list[4]
    pixel_size = raster_par_list[5]
    y_res = raster_par_list[6]
    crs = raster_par_list[7]
    
    src_ds = gdal.Open(input)
    srcband = src_ds.GetRasterBand(1)


    drv = ogr.GetDriverByName("ESRI Shapefile")
    dst_ds = drv.CreateDataSource(output)
    dst_layer = dst_ds.CreateLayer(output, srs = crs )

    fd = ogr.FieldDefn("HRU_ID_T", ogr.OFTInteger)
    dst_layer.CreateField(fd)
    dst_field = dst_layer.GetLayerDefn().GetFieldIndex("HRU_ID_T")

    gdal.Polygonize( srcband, None, dst_layer, dst_field, [], callback=None)
        
def vector_to_raster(output,input,attribute,raster_par_list,touch = "False",type = gdal.GDT_Float32):
    
    x_min = raster_par_list[0]
    x_max = raster_par_list[1]
    y_min = raster_par_list[2]
    y_max = raster_par_list[3]
    x_res = raster_par_list[4]
    pixel_size = raster_par_list[5]
    y_res = raster_par_list[6]
    crs = raster_par_list[7]
    
    source_ds = ogr.Open(input)
    source_layer = source_ds.GetLayer()

    target_ds = gdal.GetDriverByName('GTiff').Create(output, x_res, y_res, 1, type)
    target_ds.SetGeoTransform((x_min, pixel_size, 0, y_max, 0, -pixel_size))
    band = target_ds.GetRasterBand(1)
    band.SetNoDataValue(0)

    target_ds.SetProjection(crs.ExportToWkt())

    gdal.RasterizeLayer(target_ds, 
                        [1], source_layer, 
                        options = ["ALL_TOUCHED=%s"%touch, "ATTRIBUTE=%s" % attribute])
    
    return 

def array_to_raster(output,input,raster_par_list,touch = "False",type = gdal.GDT_Int32):
    
    x_min = raster_par_list[0]
    x_max = raster_par_list[1]
    y_min = raster_par_list[2]
    y_max = raster_par_list[3]
    x_res = raster_par_list[4]
    pixel_size = raster_par_list[5]
    y_res = raster_par_list[6]
    crs = raster_par_list[7]    

    target_ds = gdal.GetDriverByName('GTiff').Create(output, x_res, y_res, 1, type)
    target_ds.SetGeoTransform((x_min, pixel_size, 0, y_max, 0, -pixel_size))
    band = target_ds.GetRasterBand(1)
    band.SetNoDataValue(-9999)

    target_ds.SetProjection(crs.ExportToWkt())
    target_ds.GetRasterBand(1).WriteArray(input)
    target_ds.FlushCache()
    return 

def gdal_slope_raster(input,output):
    gdal.DEMProcessing(output,input, 'slope')

def gdal_aspect_raster(input,output):
    
    #This command outputs a 32-bit float raster with values 
    #between 0° and 360° representing the azimuth that slopes are facing. 
    #The definition of the azimuth is such that : 
    #0° means that the slope is facing the North, 
    #90° it’s facing the East, 180° it’s facing the South and 
    #270° it’s facing the West (provided that the top of your input raster is north oriented). 
    #The aspect value -9999 is used as the nodata value to indicate undefined 
    #aspect in flat areas with slope=0.
    
    gdal.DEMProcessing(output,input, 'aspect')
                
def proj_clip_raster(input,output,dst_src,cutlineDSName= '#'):
    
    input_raster = gdal.Open(input)
    output_raster = output
    if cutlineDSName == '#':
        warp = gdal.Warp(output_raster,input_raster,dstSRS=dst_src)    
    else:
        warp = gdal.Warp(output_raster, 
                    input_raster, 
                    cutlineDSName=cutlineDSName,
                    cropToCutline=True,
                    dstNodata = -9999)



def RasterHRUUnionInt32(OutputFolder,tempfolder,Merge_layer_shp_list,
                        Merge_ID_list,Sub_Lake_HRU_Layer,Sub_ID,
                        Landuse_ID,Soil_ID,Veg_ID,Other_Ply_ID_1,
                        Other_Ply_ID_2,pixel_size,lakehruinfo):
                        
    path_to_sub_lake_raster = os.path.join(tempfolder,'sub_lake.tif')
    path_to_landuse_raster = os.path.join(tempfolder,'landuse.tif')
    path_to_soil_raster = os.path.join(tempfolder,'soil.tif')
    path_to_veg_raster = os.path.join(tempfolder,'veg.tif')
    path_to_other1_raster = os.path.join(tempfolder,'o1.tif')
    path_to_other2_raster = os.path.join(tempfolder,'o2.tif')
    path_to_hru_temp_raster = os.path.join(tempfolder,'HRU.tif')
    path_to_lake_mask_raster = os.path.join(tempfolder,'lake_hru.tif')
    path_to_land_mask_raster = os.path.join(tempfolder,'land_hru.tif')
    path_to_landhru_shp= os.path.join(tempfolder,'land_hru.shp')
    path_to_lakehru_shp= os.path.join(tempfolder,'lake_hru.shp')
    path_to_landuse_shp = os.path.join(tempfolder,'landuse.shp')
    path_to_soil_shp = os.path.join(tempfolder,'soil.shp')
    path_to_veg_shp = os.path.join(tempfolder,'veg.shp')
    path_to_other1_shp = os.path.join(tempfolder,'o1.shp')
    path_to_other2_shp = os.path.join(tempfolder,'o2.shp')
    path_to_hru_temp_shp = os.path.join(tempfolder,'HRU.shp')

    raster_value_multi = {
    #                   2147483647,
    Sub_ID:                1000000,
    Landuse_ID:              10000,
    Soil_ID:                   100,
    Other_Ply_ID_1:              1,
    Veg_ID:                      1,
    Other_Ply_ID_2:              1,
    }
 
     #          12345678910
     #          0007422000,
     #          2147483647,
 #    'sub':       1000000,
 #    'landuse':     10000,
 #    'soil':          100,
 #    'o1':              1,
 
 
    # determine raster parameters        
    vector_fn = Sub_Lake_HRU_Layer
    source_ds = ogr.Open(vector_fn)
    source_layer = source_ds.GetLayer()
    x_min, x_max, y_min, y_max = source_layer.GetExtent()
    ras_crs = source_layer.GetSpatialRef()
    x_res = int((x_max - x_min) / pixel_size)
    y_res = int((y_max - y_min) / pixel_size)
    raster_par_list = [x_min, x_max, y_min, y_max,x_res,pixel_size,y_res,ras_crs]

    # rasterize 
    vector_to_raster(path_to_sub_lake_raster,Sub_Lake_HRU_Layer,'HRULake_ID',raster_par_list,touch='False')
               
    
    dissolve_filedname_list = ["HRULake_ID"]
    
    sub_raster = gdal.Open(path_to_sub_lake_raster)
    sub_raster_a = np.array(sub_raster.GetRasterBand(1).ReadAsArray())
    
    #check if some subbasin is missing 
    raster_unique_id = np.unique(sub_raster_a)
    raster_unique_id = raster_unique_id[raster_unique_id > 0]
    vector_unique_id = np.unique(lakehruinfo['HRULake_ID'].values)
    vector_unique_id = vector_unique_id[vector_unique_id > 0]
    mask = np.isin(vector_unique_id,raster_unique_id)
    
    if len(vector_unique_id[~mask]) > 0:
        print("Please increase the pixel_size parameter")
        print("The following element in path_output_folder/finalcat_hru_lake_info.shp is lost ")
        print(vector_unique_id[~mask])
        sys.exit()
        
    
    HRU_shape = sub_raster_a.shape
    HRU_a = sub_raster_a * raster_value_multi[Sub_ID]
    for i in range(0,len(Merge_layer_shp_list)):
        merge_shpfile = Merge_layer_shp_list[i]
        ID_shpfile = Merge_ID_list[i]
        out_raster = os.path.join(tempfolder,"%s_ras_layer.tif"%(ID_shpfile))
        vector_to_raster(out_raster,merge_shpfile,ID_shpfile,raster_par_list,touch='False')

        raster_layer_i = gdal.Open(out_raster)
        raster_layer_i_a = np.array(raster_layer_i.GetRasterBand(1).ReadAsArray())
        if raster_layer_i_a.shape[0] == HRU_shape[0] and raster_layer_i_a.shape[1] == HRU_shape[1]:
            HRU_a = HRU_a + raster_layer_i_a * raster_value_multi[ID_shpfile]
        else:
            print(ID_shpfile,"polygon may not fully cover the subbasin domain ",land_raster_a.shape)
            sys.exit()
        

    HRU_a[HRU_a < raster_value_multi[Sub_ID]] = -9999
    array_to_raster(path_to_hru_temp_raster,HRU_a,raster_par_list,touch = "False",type = gdal.GDT_Int32)
    
    raster_to_vector(path_to_hru_temp_shp,path_to_hru_temp_raster,raster_par_list)

    print("union done")

    return path_to_hru_temp_shp


import os


def preprocessing_obs_point(
    mask,
    path_obsin_in="#",
    obs_attributes=[],
    grassdb="#",
    grass_location="#",
    qgis_prefix_path="#",
    gis_platform="qgis",
    obsname="obs",
):

    if gis_platform == "qgis":
        assert (
            grassdb != "#"
        ), "grass database folder is needed, when gis_platform = qgis"
        assert (
            grass_location != "#"
        ), "grass location name is needed, when gis_platform = qgis"
        assert (
            qgis_prefix_path != "#"
        ), "qgis prefix path is needed, when gis_platform = qgis"
        from basinmaker.preprocessing.preinputpolygonqgis import preprocessing_input_polygon
        from basinmaker.preprocessing.reprojectandclipvectorbyplyqgis import (
            obtain_polygon_boundary,
        )
        from basinmaker.preprocessing.rasterizevectorsandloadtodbqgis import (
            rasterize_vectors_and_load_to_db,
        )
        from basinmaker.func.grassgis import grass_raster_v_import

        preprocessing_input_polygon(
            grassdb=grassdb,
            grass_location=grass_location,
            qgis_prefix_path=qgis_prefix_path,
            mask=mask,
            path_polygon=path_obsin_in,
            ply_name=obsname,
            attribute_names=[obs_attributes[0]],
            raster_names=[obsname],
        )



from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *


def define_project_extent_using_outlet_point(
    grassdb,
    grass_location,
    qgis_prefix_path,
    path_dem_in,
    outlet_pt,
    buffer_distance=0,
    mask="MASK",
    dem="dem",
):

    """Define processing extent

    Function that used to define project processing spatial extent (PSE).
    The processing spatial extent is a region where Toolbox will work in. Toolbox
    will not process grids or features outside the processing spatial extent.
    Several options is available here. The PSE
    can be defined using DEM and an point coordinates. the drainage area
    contribute to that point coordinate will be used as boundary polygon.

    Parameters
    ----------
    grassdb                           : path (required)
        It is a path to project grass database folder
    grass_location                    : string (required)
        It is a string of grass location name
    qgis_prefix_path                  : string (required)
        It is a string of qgis prefix path
    path_dem_in                      : string (required)
        It is the path to input dem
    outlet_pt                        : list (required)
        It is list that indicate the outlet coordinates of the
        region of interest. If it is provided, the PSE
        will be defined as the drainage area controlled by this point.
    mask                             : string (optional)
        It is a output mask name, which will stored in grass_location in both
        vector and raster format
    dem                              : string (optional)
        It is a output dem raster name, which will be stored in grass_location

    Notes
    -------
    Outputs are following files

    MASK                   : raster
        it is a mask raster stored in grass database, which indicate
        the PSE. The grass database is located at
        os.path.join(grassdb, grass_location)
    dem                   : raster
        it is a dem raster stored in grass database, which is
        has the same extent with MASK. The grass database is located at
        os.path.join(grassdb, grass_location)

    Returns:
    -------
       None

    Examples
    -------
    """

    print("mask region:   using outlet point ")

    QgsApplication.setPrefixPath(qgis_prefix_path, True)
    Qgs = QgsApplication([], False)
    Qgs.initQgis()
    from processing.core.Processing import Processing
    from processing.tools import dataobjects
    from qgis import processing

    feedback = QgsProcessingFeedback()
    Processing.initialize()
    QgsApplication.processingRegistry().addProvider(QgsNativeAlgorithms())
    context = dataobjects.createContext()
    context.setInvalidGeometryCheck(QgsFeatureRequest.GeometryNoCheck)

    import grass.script as grass
    import grass.script.setup as gsetup
    from grass.pygrass.modules import Module
    from grass.pygrass.modules.shortcuts import general as g
    from grass.pygrass.modules.shortcuts import raster as r
    from grass.script import array as garray
    from grass.script import core as gcore
    from grass_session import Session

    os.environ.update(
        dict(
            GRASS_COMPRESS_NULLS="1",
            GRASS_COMPRESSOR="ZSTD",
            GRASS_VERBOSE="1",
        )
    )

    PERMANENT_Temp1 = Session()
    PERMANENT_Temp1.open(
        gisdb=grassdb,
        location=grass_location + "t1",
        create_opts="EPSG:4326",
    )

    # import clipped dem to target location
    grass_raster_r_in_gdal(
        grass=grass,
        raster_path=path_dem_in,
        output_nm=dem,
        location=grass_location + "t2",
    )

    PERMANENT_Temp1.close()

    # open location where dem is loaded, generated watershed boundary in this location
    PERMANENT_Temp = Session()
    PERMANENT_Temp.open(
        gisdb=grassdb,
        location=grass_location + "t2",
        create_opts="",
    )

    # define mask of current working enviroments
    grass_raster_r_mask(grass, dem)
    # define processing extent of the current working enviroment
    grass_raster_g_region(grass, dem)

    # generate watrshed boundary for this outlet point

    # define flow direction from dem
    grass_raster_r_watershed(
        grass,
        elevation=dem,
        drainage="dir_grass",
        accumulation="acc_grass2",
        flags="sa",
    )
    # generate watershed boundary
    grass_raster_r_water_outlet(
        grass,
        input_dir_nm="dir_grass",
        output_watshed_nm="wat_mask",
        outlet_coordinates=outlet_pt,
    )
    # define mask with watershed boundary
    grass_raster_r_mask(grass, "wat_mask")

    # export generated mask to folder ouside grass work env in tif format
    grass_raster_r_out_gdal(
        grass,
        input_nm="MASK",
        output=os.path.join(grassdb, mask + "_t1.tif"),
        format="GTiff",
    )

    # polygonize exported mask raster, buffer, and use it clip input dem
    qgis_raster_gdal_polygonize(
        processing,
        context,
        INPUT=os.path.join(grassdb, mask + "_t1.tif"),
        OUTPUT=os.path.join(grassdb, mask + "_t1.shp"),
    )
    qgis_vector_dissolve(
        processing,
        context,
        INPUT=os.path.join(grassdb, mask + "_t1.shp"),
        FIELD="DN",
        OUTPUT=os.path.join(grassdb, mask + "_t2.shp"),
    )
    qgis_vector_buffer(
        processing,
        context,
        INPUT=os.path.join(grassdb, mask + "_t2.shp"),
        Buffer_Distance=buffer_distance,
        OUTPUT=os.path.join(grassdb, mask + ".shp"),
    )
    # using saga function, becasue for some reason the gdal function did
    # not reduce the extent of the raster, the raster still has the same
    # size, but the value out side of mask is set to null
    qgis_raster_saga_clip_raster_with_polygon(
        processing,
        context,
        Input=path_dem_in,
        MASK=os.path.join(grassdb, mask + ".shp"),
        Output=os.path.join(grassdb, dem + ".sdat"),
    )

    # load clipted dem raster to target location
    grass_raster_r_in_gdal(
        grass=grass,
        raster_path=os.path.join(grassdb, dem + ".sdat"),
        output_nm=dem,
        location=grass_location,
    )

    PERMANENT_Temp.close()

    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")

    # define mask of current working enviroments
    grass_raster_r_mask(grass, dem)
    # define processing extent of the current working enviroment
    grass_raster_g_region(grass, dem)

    grass_raster_r_out_gdal(
        grass,
        input_nm=mask,
        output=os.path.join(grassdb, mask + ".tif"),
        format="GTiff",
    )
    PERMANENT.close()
    Qgs.exit()



from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import sqlite3


def add_lake_attributes(
    grassdb,
    grass_location,
    qgis_prefix_path,
    input_geo_names,
    lake_attributes,
    catinfo,
):
    catchments = input_geo_names["catchment_without_merging_lakes"]
    sl_connected_lake = input_geo_names["sl_connected_lake"]
    sl_non_connected_lake = input_geo_names["sl_nonconnect_lake"]
    outlet_pt_info = input_geo_names["outlet_pt_info"]
    all_lakes = input_geo_names["all_lakes"]
    lake_outflow_pourpoints = input_geo_names["lake_outflow_pourpoints"]

    import grass.script as grass
    import grass.script.setup as gsetup
    from grass.pygrass.modules import Module
    from grass.pygrass.modules.shortcuts import general as g
    from grass.pygrass.modules.shortcuts import raster as r
    from grass.script import array as garray
    from grass.script import core as gcore
    from grass_session import Session

    os.environ.update(
        dict(GRASS_COMPRESS_NULLS="1", GRASS_COMPRESSOR="ZSTD", GRASS_VERBOSE="1")
    )
    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")

    con = sqlite3.connect(
        os.path.join(grassdb, grass_location, "PERMANENT", "sqlite", "sqlite.db")
    )

    # keep large lake ids
    grass.run_command(
        "r.stats.zonal",
        base=catchments,
        cover=sl_connected_lake,
        method="min",
        output="Connect_Lake_Cat_w_Lake_ID",
        overwrite=True,
    )

    grass.run_command(
        "v.what.rast",
        map=outlet_pt_info,
        raster=sl_non_connected_lake,
        column="ncl",
    )

    grass.run_command(
        "v.what.rast",
        map=outlet_pt_info,
        raster="Connect_Lake_Cat_w_Lake_ID",
        column="cl",
    )

    grass.run_command(
        "v.db.addcolumn", map=lake_outflow_pourpoints, columns="lakeid int"
    )

    grass.run_command(
        "v.db.update", map=lake_outflow_pourpoints, column="lakeid", qcol="cat"
    )

    grass.run_command(
        "v.what.rast",
        map=lake_outflow_pourpoints,
        raster=catchments,
        column="SubId",
    )

    ### read catchment
    sqlstat = "SELECT SubId,lakeid FROM %s" % (lake_outflow_pourpoints)
    lakeoutinfo = pd.read_sql_query(sqlstat, con)
    lakeoutinfo = lakeoutinfo.fillna(-9999)

    ### read catchment
    sqlstat = "SELECT SubId,DowSubId,ncl,cl FROM %s" % (outlet_pt_info)
    outletinfo = pd.read_sql_query(sqlstat, con)
    outletinfo = outletinfo.fillna(-9999)
    lakeinfo = Dbf_To_Dataframe(os.path.join(grassdb, all_lakes + ".shp"))

    for i in range(0, len(outletinfo)):
        catid = outletinfo["SubId"].values[i]
        catrow = catinfo["SubId"] == catid

        CL_LakeId_Outlet = outletinfo["cl"].values[i]
        ### catchment can be an connect lake catchment only when outlet of this catchment is
        ### in the lake catchment
        if CL_LakeId_Outlet > 0:
            CL_LakeId = CL_LakeId_Outlet
        else:
            CL_LakeId = -9999

        NCL_LakeId = outletinfo["ncl"].values[i]
        #        print(CL_LakeId,NCL_LakeId)
        ### add lake info
        if CL_LakeId > 0 and NCL_LakeId < 0:
            catinfo.loc[catrow, "Lake_Cat"] = 1
            catinfo.loc[catrow, "HyLakeId"] = CL_LakeId
            slakeinfo = lakeinfo.loc[lakeinfo[lake_attributes[0]] == CL_LakeId]
            if len(slakeinfo) > 0:
                catinfo.loc[catrow, "LakeVol"] = slakeinfo.iloc[0][lake_attributes[3]]
                catinfo.loc[catrow, "LakeArea"] = slakeinfo.iloc[0][lake_attributes[2]]*1000*1000
                catinfo.loc[catrow, "LakeDepth"] = slakeinfo.iloc[0][lake_attributes[4]]
                catinfo.loc[catrow, "Laketype"] = slakeinfo.iloc[0][lake_attributes[1]]
        if NCL_LakeId > 0 and CL_LakeId < 0:
            catinfo.loc[catrow, "Lake_Cat"] = 2
            catinfo.loc[catrow, "HyLakeId"] = NCL_LakeId
            slakeinfo = lakeinfo.loc[lakeinfo[lake_attributes[0]] == NCL_LakeId]
            if len(slakeinfo) > 0:
                catinfo.loc[catrow, "LakeVol"] = slakeinfo.iloc[0][lake_attributes[3]]
                catinfo.loc[catrow, "LakeArea"] = slakeinfo.iloc[0][lake_attributes[2]]*1000*1000
                catinfo.loc[catrow, "LakeDepth"] = slakeinfo.iloc[0][lake_attributes[4]]
                catinfo.loc[catrow, "Laketype"] = slakeinfo.iloc[0][lake_attributes[1]]
        if NCL_LakeId > 0 and CL_LakeId > 0:
            if catinfo["RivLength"].values[catrow] <= 0:
                catinfo.loc[catrow, "Lake_Cat"] = 2
                catinfo.loc[catrow, "HyLakeId"] = NCL_LakeId
                slakeinfo = lakeinfo.loc[lakeinfo[lake_attributes[0]] == NCL_LakeId]
                if len(slakeinfo) > 0:
                    catinfo.loc[catrow, "LakeVol"] = slakeinfo.iloc[0][lake_attributes[3]]
                    catinfo.loc[catrow, "LakeArea"] = slakeinfo.iloc[0][lake_attributes[2]]*1000*1000
                    catinfo.loc[catrow, "LakeDepth"] = slakeinfo.iloc[0][lake_attributes[4]]
                    catinfo.loc[catrow, "Laketype"] = slakeinfo.iloc[0][lake_attributes[1]]
            else:
                catinfo.loc[catrow, "Lake_Cat"] = 1
                catinfo.loc[catrow, "HyLakeId"] = CL_LakeId
                slakeinfo = lakeinfo.loc[lakeinfo[lake_attributes[0]] == CL_LakeId]
                if len(slakeinfo) > 0:
                    catinfo.loc[catrow, "LakeVol"] = slakeinfo.iloc[0][lake_attributes[3]]
                    catinfo.loc[catrow, "LakeArea"] = slakeinfo.iloc[0][lake_attributes[2]]*1000*1000
                    catinfo.loc[catrow, "LakeDepth"] = slakeinfo.iloc[0][lake_attributes[4]]
                    catinfo.loc[catrow, "Laketype"] = slakeinfo.iloc[0][lake_attributes[1]]

        #  #check if down subid of the lake cl lake catchments
        # if CL_LakeId > 0:
        #     downsubid = catinfo["DowSubId"].values[catrow][0]
        #     subid_inlakes = outletinfo.loc[outletinfo["cl"] == CL_LakeId]["SubId"].values
        #     # if downsubid is one of the lake subbasin continue, if not check if it is
        #     # the lake outlet
        #     if downsubid in subid_inlakes and downsubid != catid:
        #         continue
        #     i_lakeinfo = lakeoutinfo.loc[lakeoutinfo['lakeid'] == CL_LakeId]
        #     if len(i_lakeinfo) <=0:
        #         continue
        #     lake_outlet_subid = i_lakeinfo['SubId'].values[0]
        #
        #     # if catid not the lake outlet subid
        #     # and this cat is not flowinto the lake catchment
        #     # force it to flow into the lake outlet catchment
        #     if catid != lake_outlet_subid:
        #         catinfo.loc[catrow, "DowSubId"] = lake_outlet_subid
        #         print(CL_LakeId,catid,downsubid,lake_outlet_subid)

    PERMANENT.close()
    return catinfo



import os

class postprocess:

    def __init__(
        self,
    ):
        if os.getenv("QGIS_PREFIX_PATH"):
            self.qgispp = os.environ["QGIS_PREFIX_PATH"]
        else:
            self.qgispp = '#'

    def Generate_Raven_Model_Inputs(
        self,
        path_output_folder,
        path_hru_polygon="#",
        aspect_from_gis = "grass",
        model_name="test",
        subbasingroup_names_channel=["Allsubbasins"],
        subbasingroup_length_channel=[-1],
        subbasingroup_name_lake=["AllLakesubbasins"],
        subbasingroup_area_lake=[-1],
        lake_out_flow_method = 'broad_crest',
        model_time_step = 1,
    ):
        """This function is to generate Raven input files. A subfolder 'RavenInput'
        in path_output_folder is created to sotre resultant outputs.

        Parameters
        ----------
        path_output_folder                   : string
            is the folder path that stores generated Raven input fiels
        path_hru_polygon     : string
            is path of the output HRU shapefile from BasinMaker which includes all
            required hydrologic attributes and parameters; Each row in the attribute
            table of this shapefile represent a HRU.
        aspect_from_gis                  : string
            is a string indicating how aspect of each HRU in path_hru_polygon is
            caculated. "grass" represent the aspect is calcuated by GRASS GIS;
            "arcgis" represent the aspect is calcuated by ArcGIS.
        model_name      : string (optional)
            is the The Raven model base name.
        subbasingroup_names_channel       : list (optional)
            is a list of names for subbasin groups, which are grouped based
            on channel length of each subbsin. Should at least has one name
        subbasingroup_length_channel   : list (optional)
            is a list of channel length thresthold in meter, that divides
            subbasin into different groups. For example, [1,10,20] divides
            subbasins into four groups:

            | 1) group 1 with channel length range from (0,1];
            | 2) group 2 with channel length range from (1,10];
            | 3) group 3 with channel length range from (10,20];
            | 4) group 4 with channel length range from (20,Max channel length].
        subbasingroup_name_lake          : list (optional)
            is a list of names for subbasin groups, which are grouped based
            on area of lakes in lake subbasins. Should at least has one name
        subbasingroup_area_lake        : list (optional)
            is a list of lake area thresthold in m2, that divides
            subbasin into different groups. For example, [1,10,20] divides
            subbasins into four groups:

            | 1) group 1 with lake area range from (0,1];
            | 2) group 2 with lake are range from (1,10],
            | 3) group 3 with lake are range from (10,20],
            | 4) group 4 with lake are range from (20,max_lake_area].
        Returns
        -------

        Notes
        -------
        The following ouput files will be generated in "<path_output_folder>/RavenInput"

        | modelname.rvh              - contains subbasins and HRUs
        | Lakes.rvh                  - contains definition and parameters of lakes
        | channel_properties.rvp     - contains definition and parameters for channels

        Examples
        -------

        """

        from basinmaker.hymodin.raveninput import (
            GenerateRavenInput,
        )

        startyear=-1
        endYear=-1
        CA_HYDAT="#"
        warmup=0
        template_folder="#"
        lake_as_gauge=False
        writeobsrvt=False
        downloadobsdata=False
        forcing_input_file="#"


        GenerateRavenInput(
            Path_final_hru_info=path_hru_polygon,
            lenThres=0.2,
            iscalmanningn=1,
            Startyear=startyear,
            EndYear=endYear,
            CA_HYDAT=CA_HYDAT,
            WarmUp=warmup,
            time_step = model_time_step,
            Template_Folder=template_folder,
            Lake_As_Gauge=lake_as_gauge,
            WriteObsrvt=writeobsrvt,
            DownLoadObsData=downloadobsdata,
            Model_Name=model_name,
            Old_Product=False,
            SubBasinGroup_NM_Channel=subbasingroup_names_channel ,
            SubBasinGroup_Length_Channel=subbasingroup_length_channel ,
            SubBasinGroup_NM_Lake=subbasingroup_name_lake ,
            SubBasinGroup_Area_Lake=subbasingroup_area_lake ,
            OutputFolder=path_output_folder,
            Forcing_Input_File=forcing_input_file,
            aspect_from_gis = aspect_from_gis,
            lake_out_flow_method = lake_out_flow_method,
        )


    def obtain_grids_polygon_from_netcdf_file_method(
        self,
        netcdf_path="#",
        output_folder="#",
        coor_x_nm="lon",
        coor_y_nm="lat",
        is_rotated_grid=1,
        r_coor_x_nm="rlon",
        r_coor_y_nm="rlat",
        spatial_ref="EPSG:4326",
        x_add=-360,
        y_add=0,
        gis_platform="qgis",
    ):


        from basinmaker.postprocessing.postprocessingfunctions import (
            obtain_grids_polygon_from_netcdf_file
        )

        obtain_grids_polygon_from_netcdf_file(
            netcdf_path=netcdf_path,
            output_folder=output_folder,
            coor_x_nm=coor_x_nm,
            coor_y_nm=coor_y_nm,
            is_rotated_grid=is_rotated_grid,
            r_coor_x_nm=r_coor_x_nm,
            r_coor_y_nm=r_coor_y_nm,
            spatial_ref=spatial_ref,
            x_add=x_add,
            y_add=y_add,
            qgis_prefix_path=self.qgispp,
            gis_platform=gis_platform,
        )


    def generate_area_weight_of_two_polygons_method(
        self,
        target_polygon_path="#",
        mapping_polygon_path="#",
        col_nm="HRU_ID",
        output_folder="#",
        gis_platform='qgis',
    ):

        from basinmaker.postprocessing.postprocessingfunctions import (
            generate_area_weight_of_two_polygons
        )

        generate_area_weight_of_two_polygons(
            target_polygon_path=target_polygon_path,
            mapping_polygon_path=mapping_polygon_path,
            col_nm=col_nm,
            output_folder=output_folder,
            qgis_prefix_path=self.qgispp,
            gis_platform=gis_platform,
        )


    def Remove_Small_Lakes(
        self,
        path_output_folder="#",
        routing_product_folder  = '#',
        connected_lake_area_thresthold=-1,
        non_connected_lake_area_thresthold =-1,
        selected_lake_ids=[],
        gis_platform="qgis",
    ):
        """This function is to simplify the hydrologic routing network by removing
        lakes.

        Parameters
        ----------
        path_output_folder                   : string
            is the folder path that stores generated outputs
        routing_product_folder         : string
            is the folder path where the input hydrologic routing network is stored
        gis_platform                   : string
            is the parameter indicating which gis platform is used. It can be
            either "qgis" or "arcgis".
        connected_lake_area_thresthold                 : float (optional)
            is a lake area thresthold for connected lakes in km2.
            Connected lake with lake area below which lake will be removed
        non_connected_lake_area_thresthold             : float (optional)
            is a lake area thresthold for non-connected lakes in km2
            Non connected lake with lake area below below which lake will be removed
        selected_lake_ids          : list (optional)
            A list of lake IDs from in the hydrologic routing network (Column 'HyLakeId').
            Lakes with their lake ID in this list will be kept by the BasinMaker even
            if their area smaller than the lake area threstholds

        Returns
        -------

        Notes
        -----
        This function has no return values, The simplified hydrological routing
        network will be generated in the path_output_folder including following
        files:

        finalcat_info.shp                     : shapefile
            The finalized hydrologic routing network. the GIS layer containing
            subbasin polygons which respect the lake inflow and outflow routing
            structures. This layer contains all the necessary information for
            hydrologic routing through the lake-river network.
        finalcat_info_riv.shp                 : shapefile
            The finalized hydrologic routing network. the GIS layer containing
            river network polylines in the routing network.
        catchment_without_merging_lakes.shp             : shapefile
            The GIS layer containing subbasin polygons of an incomplete hydrologic
            routing network. In this incomplete hydrologic routing network subbasin
            polygons covered by the same lake are not merged into one lake subbasin
            yet. This incomplete hydrologic routing network is only intended as
            input to customize the routing network with our BasinMaker GIS toolbox
            (for example by defining new lake area thresholds and/or a new catchment
            minimum drainage area threshold)
        river_without_merging_lakes.shp                 : shapefile
            The GIS layer containing river polylines of an incomplete hydrologic
            routing network. In this incomplete hydrologic routing network, the
            river polylines covered by the same lake are not merged into one river
            segment yet. This incomplete hydrologic routing network is only intended as
            input to customize the routing network with our BasinMaker GIS toolbox
            (for example by defining new lake area thresholds and/or a new catchment
            minimum drainage area threshold)
        sl_connected_lake.shp                           : shapefile
            the GIS layer containing the lake polygons of lakes that are connected
            by the river_without_merging_lakes.shp
        sl_non_connected_lake.shp                       : shapefile
            the GIS layer containing the lake polygons of lakes that are not connected
            by the river_without_merging_lakes.shp
        obs_gauges                                      : shapefile
            It is the point shapefile that represent the point of interest
            after snap to river network.

        Examples
        -------


        """
        from basinmaker.postprocessing.postprocessingfunctions import (
            simplify_routing_structure_by_filter_lakes_method,
        )
        Path_final_riv_ply = '#'
        Path_final_riv = '#'
        Path_Con_Lake_ply = '#'
        Path_NonCon_Lake_ply = '#'
        simplify_routing_structure_by_filter_lakes_method(
            Path_final_riv_ply=Path_final_riv_ply,
            Path_final_riv=Path_final_riv,
            Path_Con_Lake_ply=Path_Con_Lake_ply,
            Path_NonCon_Lake_ply=Path_NonCon_Lake_ply,
            Routing_Product_Folder = routing_product_folder ,
            Thres_Area_Conn_Lakes=connected_lake_area_thresthold ,
            Thres_Area_Non_Conn_Lakes=non_connected_lake_area_thresthold ,
            Selected_Lake_List_in=selected_lake_ids ,
            OutputFolder=path_output_folder ,
            qgis_prefix_path=self.qgispp,
            gis_platform=gis_platform,
        )

        from basinmaker.postprocessing.postprocessingfunctions import (
            combine_catchments_covered_by_the_same_lake_method,
        )

        combine_catchments_covered_by_the_same_lake_method(
            Routing_Product_Folder = path_output_folder,
            qgis_prefix_path=self.qgispp,
            gis_platform=gis_platform,
        )


    def Decrease_River_Network_Resolution(
        self,
        path_output_folder="#",
        routing_product_folder = '#',
        gis_platform="qgis",
        minimum_subbasin_drainage_area=-1,
    ):
        """ This function is to simplify the hydrologic routing network by
        removing subbasins/river reaches with their drainage area below user provided
        drainage area thresthold.

        Parameters
        ----------

        Parameters
        ----------
        path_output_folder                   : string
            is the folder path that stores generated outputs
        routing_product_folder         : string
            is the folder path where the input hydrologic routing network is stored
        gis_platform                   : string
            is the parameter indicating which gis platform is used. It can be
            either "qgis" or "arcgis".
        minimum_subbasin_drainage_area                 : float
            is a subbasin drainage area thresthold, subbasin with their
            drainage area smaller than this thresthold will be removed.


        Returns
        -------

        Notes
        -----
        This function has no return values, The simplified hydrological routing
        network will be generated in the path_output_folder including following
        files:

        finalcat_info.shp                     : shapefile
            The finalized hydrologic routing network. the GIS layer containing
            subbasin polygons which respect the lake inflow and outflow routing
            structures. This layer contains all the necessary information for
            hydrologic routing through the lake-river network.
        finalcat_info_riv.shp                 : shapefile
            The finalized hydrologic routing network. the GIS layer containing
            river network polylines in the routing network.
        catchment_without_merging_lakes.shp             : shapefile
            The GIS layer containing subbasin polygons of an incomplete hydrologic
            routing network. In this incomplete hydrologic routing network subbasin
            polygons covered by the same lake are not merged into one lake subbasin
            yet. This incomplete hydrologic routing network is only intended as
            input to customize the routing network with our BasinMaker GIS toolbox
            (for example by defining new lake area thresholds and/or a new catchment
            minimum drainage area threshold)
        river_without_merging_lakes.shp                 : shapefile
            The GIS layer containing river polylines of an incomplete hydrologic
            routing network. In this incomplete hydrologic routing network, the
            river polylines covered by the same lake are not merged into one river
            segment yet. This incomplete hydrologic routing network is only intended as
            input to customize the routing network with our BasinMaker GIS toolbox
            (for example by defining new lake area thresholds and/or a new catchment
            minimum drainage area threshold)
        sl_connected_lake.shp                           : shapefile
            the GIS layer containing the lake polygons of lakes that are connected
            by the river_without_merging_lakes.shp
        sl_non_connected_lake.shp                       : shapefile
            the GIS layer containing the lake polygons of lakes that are not connected
            by the river_without_merging_lakes.shp
        obs_gauges                                      : shapefile
            It is the point shapefile that represent the point of interest
            after snap to river network.

        Examples
        -------


        """
        from basinmaker.postprocessing.postprocessingfunctions import (
            simplify_routing_structure_by_drainage_area_method,
        )

        simplify_routing_structure_by_drainage_area_method(
            Routing_Product_Folder = routing_product_folder,
            Area_Min=minimum_subbasin_drainage_area,
            OutputFolder=path_output_folder,
            gis_platform=gis_platform,
            qgis_prefix_path=self.qgispp,
        )

        from basinmaker.postprocessing.postprocessingfunctions import (
            combine_catchments_covered_by_the_same_lake_method,
        )

        combine_catchments_covered_by_the_same_lake_method(
            Routing_Product_Folder = path_output_folder,
            qgis_prefix_path=self.qgispp,
            gis_platform=gis_platform,
        )

    def Select_Subregion_Of_Routing_Structure(
        self,
        path_output_folder="#",
        routing_product_folder = '#',
        gis_platform="qgis",
        most_down_stream_subbasin_ids=[-1],
        most_up_stream_subbasin_ids=[-1],
    ):
        """Select subregion of hydrologic routing network based on provided subbasin IDs

        Parameters
        ----------
        path_output_folder                   : string
            is the folder path that stores generated outputs
        routing_product_folder         : string
            is the folder path where the input hydrologic routing network is stored
        gis_platform                   : string
            is the parameter indicating which gis platform is used. It can be
            either "qgis" or "arcgis".
        most_down_stream_subbasin_ids                    : list
            A list of subbasin ID, the subbasin IDs in this list should
            be the most downstream subbasin ID of each interested watershed.
        most_up_stream_subbasin_ids                      : list
            A list of subbasin ID, the subbasin IDs in this list should be
            the most upstream subbasin ID of each interested watershed.
            Value -1 is required to extracting the entire contribute subbains.
            subbasins between most_down_stream_subbasin_ids[i] and
            most_up_stream_subbasin_ids[i] are extracted when positive
            most_up_stream_subbasin_ids[i] is provided. i represent ith time in
            most_down_stream_subbasin_ids and  most_up_stream_subbasin_ids

        Returns
        -------

        Notes
        -----
        This function has no return values, The extracted hydrological routing
        network will be generated in the path_output_folder including following
        files:

        finalcat_info.shp                     : shapefile
            The finalized hydrologic routing network. the GIS layer containing
            subbasin polygons which respect the lake inflow and outflow routing
            structures. This layer contains all the necessary information for
            hydrologic routing through the lake-river network.
        finalcat_info_riv.shp                 : shapefile
            The finalized hydrologic routing network. the GIS layer containing
            river network polylines in the routing network.
        catchment_without_merging_lakes.shp             : shapefile
            The GIS layer containing subbasin polygons of an incomplete hydrologic
            routing network. In this incomplete hydrologic routing network subbasin
            polygons covered by the same lake are not merged into one lake subbasin
            yet. This incomplete hydrologic routing network is only intended as
            input to customize the routing network with our BasinMaker GIS toolbox
            (for example by defining new lake area thresholds and/or a new catchment
            minimum drainage area threshold)
        river_without_merging_lakes.shp                 : shapefile
            The GIS layer containing river polylines of an incomplete hydrologic
            routing network. In this incomplete hydrologic routing network, the
            river polylines covered by the same lake are not merged into one river
            segment yet. This incomplete hydrologic routing network is only intended as
            input to customize the routing network with our BasinMaker GIS toolbox
            (for example by defining new lake area thresholds and/or a new catchment
            minimum drainage area threshold)
        sl_connected_lake.shp                           : shapefile
            the GIS layer containing the lake polygons of lakes that are connected
            by the river_without_merging_lakes.shp
        sl_non_connected_lake.shp                       : shapefile
            the GIS layer containing the lake polygons of lakes that are not connected
            by the river_without_merging_lakes.shp
        obs_gauges                                      : shapefile
            It is the point shapefile that represent the point of interest
            after snap to river network.

        Examples
        -------

        """
        from basinmaker.postprocessing.postprocessingfunctions import (
            select_part_of_routing_product_method,
        )
        Path_Points="#"
        Gauge_NMS=["#"]
        Path_Catchment_Polygon="#"
        Path_River_Polyline="#"
        Path_Con_Lake_ply="#"
        Path_NonCon_Lake_ply="#"

        select_part_of_routing_product_method(
            Path_Points=Path_Points,
            Gauge_NMS=Gauge_NMS,
            OutputFolder=path_output_folder,
            mostdownid=most_down_stream_subbasin_ids,
            mostupid=most_up_stream_subbasin_ids,
            Path_Catchment_Polygon=Path_Catchment_Polygon,
            Path_River_Polyline=Path_River_Polyline,
            Path_Con_Lake_ply=Path_Con_Lake_ply,
            Path_NonCon_Lake_ply=Path_NonCon_Lake_ply,
            qgis_prefix_path=self.qgispp,
            Routing_Product_Folder = routing_product_folder,
            gis_platform=gis_platform,
        )

        from basinmaker.postprocessing.postprocessingfunctions import (
            combine_catchments_covered_by_the_same_lake_method,
        )

        combine_catchments_covered_by_the_same_lake_method(
            Routing_Product_Folder = path_output_folder,
            qgis_prefix_path=self.qgispp,
            gis_platform=gis_platform,
        )

    def Generate_HRUs(
        self,
        path_output_folder,
        gis_platform,
        path_subbasin_polygon,
        path_landuse_info,
        path_soil_info,
        path_veg_info,
        projected_epsg_code='EPSG:3573',
        path_connect_lake_polygon="#",
        path_non_connect_lake_polygon="#",
        path_landuse_polygon="#",
        path_soil_polygon="#",
        path_vegetation_polygon="#",
        path_other_polygon_1="#",
        path_other_polygon_2="#",
        importance_order = [],
        min_hru_subbasin_area_ratio = 0.0,
        path_to_dem="#",
        pixel_size = 30,
        area_ratio_thresholds = [0,0,0]
    ):
        """This function is to generate HRU map based on subbasin polygon, lake
        polygon (optional), Land use polygon (optional), soil type polygon(optional),
        vegetation polygon (optional), and two other user defined polygons
        (optional).

        Parameters
        ----------
        path_output_folder                   : string
            is the folder path that stores generated outputs
        gis_platform                   : string
            is the parameter indicating which gis platform is used. It can be
            either "qgis" or "arcgis".
        path_subbasin_polygon                 : string
            is the path of the subbasin polygon, which is generated by
            BasinMaker.
        path_landuse_info                      : string
            Path to a csv file that contains landuse information, including
            the following attributes:

            | Landuse_ID (integer) -- the landuse ID in the landuse polygon,-1 for lake
            | LAND_USE_C (string) -- the landuse class name for each landuse type
        path_soil_info                        : string
            is the path to a csv file that contains soil information, including
            following attributes:

            | Soil_ID (integer) -- the soil ID  in the soil polygon,-1 for lake
            | SOIL_PROF (string) -- the soil profile name for each soil profile type
        path_veg_info                         : string
            is the path to a csv file that contains vegetation information, including
            following attributes:

            | Veg_ID (integer) -- the vegetation ID  in the vegetation polygon,-1 for lake
            | VEG_C (string) -- the vegetation class name for each vegetation Type
        projected_epsg_code                     : string (optional)
            is a EPSG code to indicate a projected coordinate system.
            If the routing network generated by basinmaker is under a
            projected system, please assgin the EPSG code of that projected
            system to projected_epsg_code. If the routing network generated by
            basinmaker is under a geographic system, please use EPSG code from
            any projected system. The coordinate system of routing network from
            basinmaker is determined by the coordinate system  of input DEM.
        path_connect_lake_polygon            : string (Optional)
            is the path to the connected lake's polygon
        path_non_connect_lake_polygon        : string (Optional)
            is the path to the non connected lake's polygon
        path_landuse_polygon                 : string (Optional)
            is the path to the landuse polygon. when path_landuse_polygon is not
            provided. The Landuse ID in path_landuse_info should be
            1: land, -1: lake
        path_soil_polygon                    : string (Optional)
            is the path to the soil polygon. when soil polygon is not
            provided. The Soil ID in path_soil_info should be the same
            as Landuse ID.
        path_vegetation_polygon                     : string (Optional)
            is the path to the vegetation polygon. when vegetation polygon is not
            provided. The Veg ID in path_veg_info should be the same
            as Landuse ID.
        path_other_polygon_1                 : string (Optional)
            is the path to the other polygon that will be used to define HRU,
            such as elevation band, or aspect.
        path_other_polygon_2                 : string (Optional)
            is the path to the other polygon that will be used to define HRU,
            such as elevation band, or aspect.
        DEM                              : string (optional)
            is the path to a raster elevation dataset, that is used to
            calcuate average apspect, elevation and slope within each HRU.
            if no data is provided, subbasin averaged value will be used for
            each HRU.
        Returns
        -------

        Notes
        -----
        This function has no return values, but a HRU map saved in the
        path_output_folder

        Examples
        -------

        """
        from basinmaker.postprocessing.postprocessingfunctions import (
            generate_hrus_method,
        )
        Sub_Lake_ID="HyLakeId"
        Sub_ID="SubId"
        Lake_Id="Hylak_id"
        Landuse_ID="Landuse_ID"
        Soil_ID="Soil_ID"
        Other_Ply_ID_1="O_ID_1"
        Veg_ID="Veg_ID"
        Other_Ply_ID_2="O_ID_2"

        os.environ["GISDBASE"] = path_output_folder

        generate_hrus_method(
            Path_Subbasin_Ply=path_subbasin_polygon,
            Landuse_info=path_landuse_info,
            Soil_info=path_soil_info,
            Veg_info=path_veg_info,
            Sub_Lake_ID=Sub_Lake_ID,
            Sub_ID=Sub_ID,
            Path_Connect_Lake_ply=path_connect_lake_polygon,
            Path_Non_Connect_Lake_ply=path_non_connect_lake_polygon,
            Lake_Id=Lake_Id,
            Path_Landuse_Ply=path_landuse_polygon,
            Landuse_ID=Landuse_ID,
            Path_Soil_Ply=path_soil_polygon,
            Soil_ID=Soil_ID,
            Path_Veg_Ply=path_vegetation_polygon,
            Veg_ID=Veg_ID,
            Path_Other_Ply_1=path_other_polygon_1,
            Other_Ply_ID_1=Other_Ply_ID_1,
            Path_Other_Ply_2=path_other_polygon_2,
            Other_Ply_ID_2=Other_Ply_ID_2,
            DEM=path_to_dem,
            Inmportance_order = importance_order,
            min_hru_area_pct_sub = min_hru_subbasin_area_ratio,
            Project_crs=projected_epsg_code,
            OutputFolder=path_output_folder,
            qgis_prefix_path=self.qgispp,
            gis_platform = gis_platform,
            pixel_size = pixel_size,
            area_ratio_thresholds = area_ratio_thresholds,
        )


class delineate:

    """
    QGIS/GRASSToolsets to delelineate lake river routing structure
    ...

    Attributes
    ----------


    Methods
    -------

    """

    def __init__(
        self,
        path_output_folder="#",
        path_working_folder="#",
    ):

        # define drived values
        # create folders
        self.path_output_folder = path_output_folder
        self.path_working_folder = path_working_folder

        #        os.makedirs(self.path_output_folder, exist_ok=True)
        os.makedirs(self.path_working_folder, exist_ok=True)

        # obtain qgis prefix path
        if os.getenv("QGIS_PREFIX_PATH"):
            self.qgispp = os.environ["QGIS_PREFIX_PATH"]
        else:
            self.qgispp = '#'
        # obtain basinmaker path
        # define grass database folder
        self.grassdb = os.path.join(self.path_working_folder, "grassdb")
        if not os.path.exists(self.grassdb):
            os.makedirs(self.grassdb)

        # add grass database folder into enviroment variable
        os.environ["GISDBASE"] = self.grassdb

        # define grass location names
        self.grass_location_geo = "main_working_location"

        # grass sql databse folder
        self.sqlpath = os.path.join(
            self.grassdb, "main_working_location", "PERMANENT", "sqlite", "sqlite.db"
        )

        # define constants

        # default channel manning's coefficient
        DEFAULT_CHN = 0.035
        # minimum channel slope
        MIN_RIV_SLP = 0.00001

        # default pre processed and well prepared spatial data name
        self.geofilenames = {
            "dem": "dem",
            "mask": "MASK",
            "dem_proj": "dem_proj",
            "fdr_grass": "fdr_grass",
            "fdr_arcgis": "fdr_arcgis",
            "acc": "acc",
            "str_r": "str_r",
            "str_v": "str_v",
            "cat_no_lake": "cat_no_lake",
            "all_lakes": "all_lakes",
            "selected_lakes": "selected_lakes",
            "connect_lake": "connect_lake",
            "nonconnect_lake": "nonconnect_lake",
            "str_connected_lake": "str_connected_lake",
            "sl_nonconnect_lake": "sl_nonconnect_lake",
            "sl_connected_lake": "sl_connected_lake",
            "str_sl_connected_lake": "str_sl_connected_lake",
            "obspoint": "obspoint",
            "flood_n": "flood_n",
            "bk_wd": "bk_wd",
            "lake_boundary": "lake_boundary",
            "nfdr_arcgis": "nfdr_arcgis",
            "nfdr_grass": "nfdr_grass",
            "catchment_without_merging_lakes": "catchment_without_merging_lakes",
            "river_without_merging_lakes": "river_without_merging_lakes",
            "cat_use_default_acc": "cat_use_default_acc",
            "snapped_obs_points": "snapped_obs_points",
            "sub_reg_str_r": "sub_reg_str_r",
            "sub_reg_str_v": "sub_reg_str_v",
            "sub_reg_nfdr_grass": "sub_reg_nfdr_grass",
            "sub_reg_nfdr_arcgis": "sub_reg_nfdr_arcgis",
            "sub_reg_acc": "sub_reg_acc",
            "sub_reg_dem": "sub_reg_dem",
            "problem_seg": "problem_seg",
            "lake_outflow_pourpoints": "lake_outflow_pourpoints",
        }

    # first modulized methods
    def Define_Project_Spatial_Extent(
        self,
        mode,
        path_to_dem_input,
        watershed_outlet_coordinates = [-1, -1],
        path_to_spatial_extent_polygon="#",
        buffer_distance=0.0,
        path_to_hydrobasin_polygon="#",
        hydrobasin_id_of_watershed_outlet=-1,
        gis_platform="qgis",
    ):

        """This function is to define project spatial extent (PSE). Domain ouside
        of the PSE is not processed by BasinMaker functions.

        Parameters
        ----------
        mode                              : string (required)
            is a string indicating which to define PSE

            | 'using_dem'            : the extent of input dem is used
            | 'using_hybasin'        : the extent is defined by subbasins that are
                                       drainage to the provided watershed outlet
                                       subbasins ID in HydroBASINS product
            | 'using_outlet_pt'      : the extent is defined by the watershed
                                       generated from input dem and
                                       the watershed outlet coordinates
            | 'using_provided_ply'   : the extent of provided polygon is used
        path_to_dem_input            : string (required)
            is the path to input dem
        watershed_outlet_coordinates : list (optional)
            is a list that indicate the outlet coordinates of the
            region of interest in [lat, lon]. It is needed when mode = 'using_outlet_pt'.
        path_to_spatial_extent_polygon                  : string (optional)
            is the path of a polygon shapefile, the extent of which will be used
            as PSE.
        buffer_distance                  : float (optional)
            is a float number to enlarge the PSE. It is needed when mode = 'using_hybasin'
            or mode = 'using_provided_ply'. It is the distance around the PSE from
            HydroBASINS or provided polygons that will be buffered. The unit is the
            same with the spatial unit of input DEM.
        path_to_hydrobasin_polygon                      : string (optional)
            is a path to the HydroBASINs product. It is needed when mode = 'using_hybasin'
        hydrobasin_id_of_watershed_outlet               : int (optional)
            is a HydroBASINS subbasin ID of the watershed outlet. It is needed
            when mode = 'using_hybasin'

        Returns
        -------

        Notes
        -------
        Outputs are following files in GRASS GIS database loacated in
        os.path.join(path_working_folder,'grassdb')

        MASK.*        : raster/shp
            it is a mask raster stored in grass database, which indicate the PSE.
        dem.*         : raster/shp
            it is a dem raster stored in grass database, which is has the same extent
            with MASK.

        Examples
        -------
        """

        from basinmaker.extent.projectextent import define_project_extent
        up_hybasin_id = -1
        define_project_extent(
            grassdb=self.grassdb,
            grass_location=self.grass_location_geo,
            qgis_prefix_path=self.qgispp,
            mode=mode,
            path_dem_in=path_to_dem_input,
            outlet_pt=watershed_outlet_coordinates,
            path_extent_ply=path_to_spatial_extent_polygon,
            buffer_distance=buffer_distance,
            hybasin_ply=path_to_hydrobasin_polygon,
            down_hybasin_id=hydrobasin_id_of_watershed_outlet,
            up_hybasin_id=up_hybasin_id,
            mask=self.geofilenames["mask"],
            dem=self.geofilenames["dem"],
            gis_platform=gis_platform,
        )

    def Delineation_Initial_Subbasins_Without_Lakes(
        self,
        fac_thresold,
        mode = 'using_dem',
        path_flow_dirction="#",
        path_flow_accumulation='#',
        max_memroy=1024 * 4,
        gis_platform="qgis",
        subreg_fdr_path="#",
        subreg_acc_path="#",
        subreg_str_r_path="#",
        subreg_str_v_path="#",
    ):
        """Function that used to generate a initial routing structure with
        user provied flow accumulation thresthold without considering lake.

        Parameters
        ----------
        fac_thresold       : float
            is the flow accumulation thresthold, used to determine
            subbsains and river network. Increasing of this paramter will
            increase the size of generated subbasins, reduce the number
            subbasins and reduce the number of generated stream reaches
        mode              : string (required)
            is a string indicate which dataset will be used to delineate
            watershed.

            | 'using_dem' : dem is used for initial subbasin delineation
            | 'using_fdr' : flow direction data is used for subbasin delineation
        path_flow_dirction          : string (optional)
            is a path indicating the path of flow direction input dataset
            only needed when mode = 'using_fdr'
        max_memroy        : integer
            is the maximum memeory that allow to be used in MB.

        Returns
        -------

        Notes
        -------
        Outputs are following files in GRASS GIS database loacated in
        os.path.join(path_working_folder,'grassdb')

        fdr_grass              : raster
            is a raster represent flow direction dataset, which is
            using 1 - 8 to represent different directions
        fdr_arcgis             : raster
            is a raster represent flow direction dataset, which is
            using 1,2,4,...64,128 to represent different directions
        str_v                  : vector
            is a river network in vector format
        str_r                  : raster
            is a river network in raster format
        cat_no_lake            : raster
            is the raster represent the delineated subbasins without
            considering lakes
        acc                    : raster
            is the raster represent the flow accumulation


        Examples
        -------
        """
        from basinmaker.delineationnolake.watdelineationwithoutlake import (
            watershed_delineation_without_lake,
        )
        if mode == 'using_dem':
            mode = 'usingdem'
        elif mode == 'using_sub_reg':
            mode = 'usingsubreg'
        elif mode == 'using_fac':
            mode = 'usingfac'
        else:
            mode = 'usingfdr'

        watershed_delineation_without_lake(
            mode=mode,
            input_geo_names=self.geofilenames,
            acc_thresold=fac_thresold,
            fdr_path=path_flow_dirction,
            fac_path = path_flow_accumulation,
            subreg_fdr_path=subreg_fdr_path,
            subreg_acc_path=subreg_acc_path,
            subreg_str_r_path=subreg_str_r_path,
            subreg_str_v_path=subreg_str_v_path,
            fdr_arcgis=self.geofilenames["fdr_arcgis"],
            fdr_grass=self.geofilenames["fdr_grass"],
            str_r=self.geofilenames["str_r"],
            str_v=self.geofilenames["str_v"],
            acc=self.geofilenames["acc"],
            cat_no_lake=self.geofilenames["cat_no_lake"],
            max_memroy=max_memroy,
            grassdb=self.grassdb,
            grass_location=self.grass_location_geo,
            qgis_prefix_path=self.qgispp,
            gis_platform=gis_platform,
        )

    def Add_New_Subbasin_Outlet_Points(
        self,
        path_lake_polygon="#",
        lake_attributes=[],
        connected_lake_area_thresthold=0,
        non_connected_lake_area_thresthold=0,
        only_included_lake_at_river_interction = False,
        path_point_of_interest="#",
        point_of_interest_attributes=[],
        path_sub_reg_outlets_v="#",
        path_sub_reg_lake_r="#",
        path_sub_reg_lake_bd_r="#",
        search_radius=100,
        mode="#",
        max_memroy=1024 * 4,
        gis_platform="qgis",
    ):
        """ Update the subbasin delineation result by adding lake inflow and
        outflow points and observation gauges as a new subbasin outlets. The output
        is not the final delineation result. because:

        | 1) Hydrologcial related attributes for each subbasin are not calcuated yet.
        | 2) Some lakes may cover several subbasins. The output needs to be finalized
             by combing those subbasins with the same lake as one subbasin only.


        Parameters
        ----------
        path_lake_polygon                 : string (optional)
            is a path of the lake polygon shapefile
        lake_attributes                   : list (optional)
            the columns names in the input lake polygon that indicate following
            items (mandatory). It is needed only when path_lake_polygon_in is
            provided. Columns (2-4 in following list) in lake poylon can be fill
            with any number, when these infomation is not avaiable.

            | 1) column name for the unique Id of each lake, datatype is integer
            | 2) column name for type of the lake, datatype is integer
            | 3) column name for the volume of lakes in km3, datatype is float
            | 4) column name for the average depth of lakes in m, datatype is float
            | 5) column name for the area of lakes in km2, datatype is float
        connected_lake_area_thresthold                 : float (optional)
            is a lake area thresthold for connected lakes in km2.
            Connected lake with lake area below this value will not be considerd
        non_connected_lake_area_thresthold             : float (optional)
            is a lake area thresthold for non-connected lakes in km2
            Non connected lake with lake area below this value will not be considered
        path_point_of_interest
            is a path of the point shapefile that indicate points of interest,which
            can include different observation gauges.
        point_of_interest_attributes                    : list (optional)
            the columns names in the point of interest shapefile that indicate
            following items (mandatory). It is needed only when path_point_of_interest
            is provided. Columns (2-4 in following list) in point shapefile can be
            fill with any value, when these infomation is not avaiable.

            | 1) column name for the unique Id of each observation point, datatype is integer
            | 2) column name for the unique name of each observation point, datatype is string
            | 3) column name for the drainage area of each observation point in km3, datatype is float
            | 4) column name for the source of the observation point: 'CA' for observation in canada;
                 'US' for observation in US, or any user-provided names, datatype is string
        max_memroy        : integer (optional)
            is the maximum memeory that allow to be used in MB.

        Returns
        -------

        Notes
        -------
        Output raster and vector files that will be used by next step are list as
        following. All files are saved in GRASS GIS database loacated in
        os.path.join(path_working_folder,'grassdb')

        selected_lakes                    : raster
            it is a raster represent all lakes that are selected by two lake
            area threstholds
        sl_nonconnect_lake       : raster
            it is a raster represent all non connected lakes that are selected
            by lake area threstholds
        sl_connected_lake           : raster
            it is a raster represent all connected lakes that are selected
            by lake area threstholds
        river_without_merging_lakes                         : raster/vector
            it is the updated river segment for each subbasin
        catchment_without_merging_lakes                     : raster/vector
            it is a raster represent updated subbasins after adding lake inflow
            and outflow points as new subbasin outlet.
        snapped_obs_points                                  : raster/vector
            it is a name of the point gis file represent successfully sanpped
            point of interest points


        Examples
        -------

        """
        from basinmaker.addlakeandobs.addlakeandobsintowatershed import (
            add_lakes_and_obs_into_existing_watershed_delineation,
        )

        add_lakes_and_obs_into_existing_watershed_delineation(
            input_geo_names=self.geofilenames,
            path_lakefile_in=path_lake_polygon,
            lake_attributes=lake_attributes,
            path_obsfile_in=path_point_of_interest,
            obs_attributes=point_of_interest_attributes,
            path_sub_reg_outlets_v=path_sub_reg_outlets_v,
            threshold_con_lake=connected_lake_area_thresthold,
            only_included_lake_at_river_interction = only_included_lake_at_river_interction,
            threshold_non_con_lake=non_connected_lake_area_thresthold,
            search_radius=search_radius,
            path_sub_reg_lake_r=path_sub_reg_lake_r,
            path_sub_reg_lake_bd_r=path_sub_reg_lake_bd_r,
            sl_connected_lake=self.geofilenames["sl_connected_lake"],
            sl_non_connected_lake=self.geofilenames["sl_nonconnect_lake"],
            sl_lakes=self.geofilenames["selected_lakes"],
            catchment_without_merging_lakes=self.geofilenames[
                "catchment_without_merging_lakes"
            ],
            river_without_merging_lakes=self.geofilenames[
                "river_without_merging_lakes"
            ],
            snapped_obs_points=self.geofilenames["snapped_obs_points"],
            cat_use_default_acc=self.geofilenames["cat_use_default_acc"],
            nfdr_arcgis=self.geofilenames["nfdr_arcgis"],
            nfdr_grass=self.geofilenames["nfdr_grass"],
            max_memroy=max_memroy,
            grassdb=self.grassdb,
            grass_location=self.grass_location_geo,
            qgis_prefix_path=self.qgispp,
            gis_platform=gis_platform,
        )

    def Generate_Hydrologic_Routing_Attributes(
        self,
        path_output_folder ="#",
        projected_epsg_code ="EPSG:3573",
        path_bkfwidthdepth_polyline ="#",
        bkfwd_attributes=[],
        path_k_c_zone_polygon = '#',
        k =-1,
        c =-1,
        path_landuse ="#",
        path_landuse_and_manning_n_table="#",
        lake_attributes =[],
        point_of_interest_attributes =[],
        outlet_obs_id=-1,
        path_sub_reg_outlets_v="#",
        gis_platform="qgis",
    ):
        """Calculate hydrological paramters for each subbasin.

        Parameters
        ----------
        path_output_folder                  : string
            The path to a folder to save outputs
        projected_epsg_code                     : string (optional)
            is a EPSG code to indicate a projected coordinate system.
            If the routing network generated by basinmaker is under a
            projected system, please assgin the EPSG code of that projected
            system to projected_epsg_code. If the routing network generated by
            basinmaker is under a geographic system, please use EPSG code from
            any projected system. The coordinate system of routing network from
            basinmaker is determined by the coordinate system  of input DEM.
        path_bkfwidthdepth_polyline             : string (optional)
            is a path of the polyline shapefile that contains bankfull width (w) and
            depth (d) data.Following the methodology in Andreadis et al. (2013),
            the w, and d of each subbasin can be calculated by these two equation
            w= 7.2 Q**0.5 and d=0.27Q**0.39 , respectively. And the the bankfull
            discharge Q of each subbasin can be estimated using the relationship
            between Q and the drainage area (DA in [km2]), which is the Q=k×DA**c.
            If the bankfull width and depth polyline data is provided, basinmaker
            will use it to estimate the k and c for the entire watershed.
        bkfwd_attributes               : list (optional)
            the columns names that indicate following items (mandatory).It is only
            needed with path_bkfwidthdepth is provided

            | 1) column name for the Bankfull width in m;
            | 2) column name for the Bankfull depth in m;
            | 3) column name for the annual mean discharge in m3/s;
            | 4) column name for the drainage area in km2;
        k                              : float (optional)
            the coefficient in Q=k×DA**c.if k and c is provided, the
            path_bkfwidthdepth_polyline will not be used
        c                              : float (optional)
            the coefficient in Q=k×DA**c.if k and c is provided, the
            path_bkfwidthdepth_polyline will not be used
        path_landuse                   : string (optional)
            is a path of the landuse raster.It is used to estimate the floodplain
            Manning's coefficient. Require the same projection with the DEM data
            and formatted in ".tif".
        path_landuse_and_manning_n_table              : string (optional)
            is a path of the table in '.csv' format.The table describe the
            floodplain Manning's coefficient correspond to a given landuse type.
            The table should have two columns:

            | RasterV: is the landuse value in the landuse raster for each land use type
            | MannV: is the roughness coefficient value for each landuse type.
        lake_attributes                   : list (optional)
            the columns names in the input lake polygon that indicate following
            items (mandatory). It is needed only when path_lake_polygon_in is
            provided. Columns (2-4 in following list) in lake poylon can be fill
            with any number, when these infomation is not avaiable.

            | 1) column name for the unique Id of each lake, datatype is integer
            | 2) column name for type of the lake, datatype is integer
            | 3) column name for the volume of lakes in km3, datatype is float
            | 4) column name for the average depth of lakes in m, datatype is float
            | 5) column name for the area of lakes in km2, datatype is float
        point_of_interest_attributes                    : list (optional)
            the columns names in the point of interest shapefile that indicate
            following items (mandatory). It is needed only when path_point_of_interest
            is provided. Columns (2-4 in following list) in point shapefile can be
            fill with any value, when these infomation is not avaiable.

            | 1) column name for the unique Id of each observation point, datatype is integer
            | 2) column name for the unique name of each observation point, datatype is string
            | 3) column name for the drainage area of each observation point in km3, datatype is float
            | 4) column name for the source of the observation point: 'CA' for observation in canada;
                 'US' for observation in US, or any user-provided names, datatype is string

        Returns
        -------

        Notes
        -------
        Five vector files will be generated in the output folder. these files
        can be further finalized as hydrologcial routing network by
        function "combine_catchments_covered_by_the_same_lake"
        or be used as input for BasinMaker post processint tools.
        The attributes included in each GIS file can be found in
        http://hydrology.uwaterloo.ca/basinmaker/index.html

        catchment_without_merging_lakes.shp             : shapefile
            The GIS layer containing subbasin polygons of an incomplete hydrologic
            routing network. In this incomplete hydrologic routing network subbasin
            polygons covered by the same lake are not merged into one lake subbasin
            yet. This incomplete hydrologic routing network is only intended as
            input to customize the routing network with our BasinMaker GIS toolbox
            (for example by defining new lake area thresholds and/or a new catchment
            minimum drainage area threshold)
        river_without_merging_lakes.shp                 : shapefile
            The GIS layer containing river polylines of an incomplete hydrologic
            routing network. In this incomplete hydrologic routing network, the
            river polylines covered by the same lake are not merged into one river
            segment yet. This incomplete hydrologic routing network is only intended as
            input to customize the routing network with our BasinMaker GIS toolbox
            (for example by defining new lake area thresholds and/or a new catchment
            minimum drainage area threshold)
        sl_connected_lake.shp                           : shapefile
            the GIS layer containing the lake polygons of lakes that are connected
            by the river_without_merging_lakes.shp
        sl_non_connected_lake.shp                       : shapefile
            the GIS layer containing the lake polygons of lakes that are not connected
            by the river_without_merging_lakes.shp
        obs_gauges                                      : shapefile
            It is the point shapefile that represent the point of interest
            after snap to river network.


        Examples
        -------

        """

        from basinmaker.addattributes.addattributestocatchments import add_attributes_to_catchments

        add_attributes_to_catchments(
            input_geo_names=self.geofilenames,
            path_bkfwidthdepth=path_bkfwidthdepth_polyline,
            bkfwd_attributes=bkfwd_attributes,
            path_landuse=path_landuse,
            path_landuse_info=path_landuse_and_manning_n_table,
            projection=projected_epsg_code,
            path_k_c_zone_polygon = path_k_c_zone_polygon,
            k_in=k,
            c_in=c,
            out_cat_name=self.geofilenames["catchment_without_merging_lakes"],
            out_riv_name=self.geofilenames["river_without_merging_lakes"],
            grassdb=self.grassdb,
            grass_location=self.grass_location_geo,
            qgis_prefix_path=self.qgispp,
            gis_platform=gis_platform,
            obs_attributes=point_of_interest_attributes,
            lake_attributes=lake_attributes,
            outlet_obs_id=outlet_obs_id,
            path_sub_reg_outlets_v=path_sub_reg_outlets_v,
            output_folder=path_output_folder,
        )

    def divide_domain_into_sub_regions_method(
        self,
        path_lakefile_in,
        lake_attributes,
        path_bkfwidthdepth="#",
        bkfwd_attributes="#",
        Min_Num_Domain=9,
        Max_Num_Domain=13,
        Initaial_Acc=5000,
        Delta_Acc=1000,
        CheckLakeArea=1,
        fdr_path="#",
        Acc_Thresthold_stream=500,
        max_memory=2048 * 3,
        Out_Sub_Reg_Folder="#",
        gis_platform="qgis",
    ):

        from basinmaker.subreg.defsubreg import (
            divide_domain_into_sub_regions,
        )

        divide_domain_into_sub_regions(
            input_geo_names=self.geofilenames,
            grassdb=self.grassdb,
            grass_location=self.grass_location_geo,
            qgis_prefix_path=self.qgispp,
            path_lakefile_in=path_lakefile_in,
            lake_attributes=lake_attributes,
            path_bkfwidthdepth=path_bkfwidthdepth,
            bkfwd_attributes=bkfwd_attributes,
            Min_Num_Domain=Min_Num_Domain,
            Max_Num_Domain=Max_Num_Domain,
            Initaial_Acc=Initaial_Acc,
            Delta_Acc=Delta_Acc,
            CheckLakeArea=CheckLakeArea,
            fdr_path=fdr_path,
            Acc_Thresthold_stream=Acc_Thresthold_stream,
            max_memory=max_memory,
            Out_Sub_Reg_Folder=Out_Sub_Reg_Folder,
            sub_reg_str_r=self.geofilenames["sub_reg_str_r"],
            sub_reg_str_v=self.geofilenames["sub_reg_str_v"],
            sub_reg_nfdr_grass=self.geofilenames["sub_reg_nfdr_grass"],
            sub_reg_nfdr_arcgis=self.geofilenames["sub_reg_nfdr_arcgis"],
            sub_reg_acc=self.geofilenames["sub_reg_acc"],
            sub_reg_dem=self.geofilenames["sub_reg_dem"],
            gis_platform=gis_platform,
        )

    def combine_sub_region_results_method(
        self,
        path_sub_region_info,
        sub_region_outputfolder,
        outputfolder,
        is_final_result,
        path_subregion_inlet,
        gis_platform="qgis",
        start_sub_id = 0,
        k =1,
        c = 1,
    ):

        from basinmaker.subreg.defsubreg import (
            combine_sub_region,
        )

        combine_sub_region(
            path_sub_region_info=path_sub_region_info,
            sub_region_outputfolder=sub_region_outputfolder,
            outputfolder=outputfolder,
            is_final_result=is_final_result,
            qgis_prefix_path=self.qgispp,
            path_subregion_inlet=path_subregion_inlet,
            start_sub_id = start_sub_id,
            k = k,
            c = c
        )

    def Combine_Subbasins_Covered_by_The_Same_Lake(
        self,
        routing_product_folder='#',
        gis_platform="qgis",
        area_thresthold = 10*30*30/1000/1000,
    ):
        """Finalize a incomplete hydrologic routing network by merging subbasin
        polygons that are covered by the same lake.

        Parameters
        ----------
        routing_product_folder         : string
            is the folder where the input routing product is stored
        gis_platform                   : string
            It is the parameter indicate which gis platform is used. It can be
            either "qgis" or "arcgis".
        Returns
        -------

        Notes
        -----
        This function has no return values, two vector files will be generated in
        the routing_product_folder

        finalcat_info.shp                     : shapefile
            The finalized hydrologic routing network. the GIS layer containing
            subbasin polygons which respect the lake inflow and outflow routing
            structures. This layer contains all the necessary information for
            hydrologic routing through the lake-river network.
        finalcat_info_riv.shp                 : shapefile
            The finalized hydrologic routing network. the GIS layer containing
            river network polylines in the routing network.

        Examples
        -------

        """
        from basinmaker.postprocessing.postprocessingfunctions import (
            combine_catchments_covered_by_the_same_lake_method,
        )

        combine_catchments_covered_by_the_same_lake_method(
            Routing_Product_Folder = routing_product_folder,
            qgis_prefix_path=self.qgispp,
            gis_platform=gis_platform,
            area_thresthold = area_thresthold,
        )



import shutil
import os


def define_project_extent(
    mode,
    path_dem_in,
    outlet_pt=[-1, -1],
    path_extent_ply="#",
    buffer_distance=0.0,
    hybasin_ply="#",
    down_hybasin_id=-1,
    up_hybasin_id=-1,
    mask="MASK",
    dem="dem",
    grassdb="#",
    grass_location="#",
    qgis_prefix_path="#",
    gis_platform="qgis",
):
    """Define processing extent

    Function that used to define project processing spatial extent (PSE).
    The processing spatial extent is a region where Toolbox will work in. Toolbox
    will not process grids or features outside the processing spatial extent.
    Several options is available here. 1) The PSE can be defined as the extent of
    input DEM. 2)The PSE can be defined using Hybasin product and a hydrobasin
    ID. All subbasin drainage to that hydrobasin ID will be extracted. And
    the extent of the extracted polygon will be used as PSE. 3)The PSE
    can be defined using DEM and an point coordinates. the drainage area
    contribute to that point coordinate will be used as boundary polygon. 4)
    The PSE can be defined using input polygon

    Parameters
    ----------
    grassdb                           : path (required)
        It is a path to project grass database folder
    grass_location                    : string (required)
        It is a string of grass location name
    qgis_prefix_path                  : string (required)
        It is a string of qgis prefix path
    mode                              : string (required)
        It is a string indicate which method to define project processing
        spatial extent
        'using_dem'            : the extent of input dem will be used
        'using_hybasin'        : the extent will be defined using hydrobasin
                                 product
        'using_outlet_pt'      : the extent will be defined with provided outlet
                                 point
        'using_provided_ply'   : the extent will be defined by provided polygon
    path_dem_in                      : string (required)
        It is the path to input dem
    outlet_pt                        : list (optional)
        It is list that indicate the outlet coordinates of the
        region of interest. If it is provided, the PSE
        will be defined as the drainage area controlled by this point.
    path_extent_ply                  : string (optional)
        It is the path of a subregion polygon. It is only used when the Region
        of interest is very large and the resolution of the dem is very high.
        toolbox will first divide the whole region into several small subregions.
        And then using devided subregion polygon as PSE.
    buffer_distance                  : float (optional)
        It is a float number to increase the extent of the PSE
        obtained from Hydrobasins. It is needed when input DEM is not from
        HydroSHEDS. Then the extent of the watershed will be different
        with PSE defined by HydroBASINS.
    hybasin_ply                      : string (optional)
        It is a path to hydrobasin routing product, If it is provided, the
        PSE will be based on the OutHyID and OutHyID2 and
        this HydroBASINS routing product.
    down_hybasin_id                  : int (optional)
        It is a HydroBASINS subbasin ID, which should be the ID of the most
        downstream subbasin in the region of interest.
    up_hybasin_id                    : int (optional)
        It is a HydroBASINS subbasin ID, which should be the ID of the most
        upstream subbasin in the region of interest, normally do not needed.
    mask                             : string (optional)
        It is a output mask name, which will stored in grass_location in both
        vector and raster format
    dem                              : string (optional)
        It is a output dem raster name, which will be stored in grass_location

    Notes
    -------
    Outputs are following files

    MASK                   : raster
        it is a mask raster stored in grass database, which indicate
        the PSE. The grass database is located at
        os.path.join(grassdb, grass_location)
    dem                   : raster
        it is a dem raster stored in grass database, which is
        has the same extent with MASK. The grass database is located at
        os.path.join(grassdb, grass_location)

    Returns:
    -------
       None

    Examples
    -------
    """
    shutil.rmtree(grassdb)
    if not os.path.exists(grassdb):
        os.makedirs(grassdb)

    if gis_platform == "qgis":
        assert (
            grassdb != "#"
        ), "grass database folder is needed, when gis_platform = qgis"
        assert (
            grass_location != "#"
        ), "grass location name is needed, when gis_platform = qgis"
        assert (
            qgis_prefix_path != "#"
        ), "qgis prefix path is needed, when gis_platform = qgis"
        from basinmaker.extent.usingdemqgis import define_project_extent_using_dem
        from basinmaker.extent.usingoutletpointqgis import define_project_extent_using_outlet_point
        from basinmaker.extent.usinginputplyqgis import define_project_extent_using_input_polygon
        from basinmaker.extent.usinghybasinplyqgis import define_project_extent_using_hybasin_ply
    
    if gis_platform == "arcgis":
        from basinmaker.extent.usingdemarcgis import define_project_extent_using_dem
        from basinmaker.extent.usinghybasinplyarcgis import define_project_extent_using_hybasin_ply
        
    if mode == "using_hybasin":
        assert (
            hybasin_ply != "#"
        ), "HydroBasin product is needed to define processing extent with mode = using_hybasin"
        assert (
            down_hybasin_id != "#"
        ), "The watershed outlet HydroBasin sub Id is needed to define processing extent with mode = using_hybasin"

    if mode == "using_outlet_pt":
        assert (
            outlet_pt[0] != "-1"
        ), "outlet_pt needs to be define processing extent with mode = using_outlet_pt"

    if mode == "using_provided_ply":
        assert (
            path_extent_ply != "#"
        ), "path_extent_ply needs to be defined when mode = using_provided_ply"

    if mode == "using_dem":
        define_project_extent_using_dem(
            grassdb,
            grass_location,
            qgis_prefix_path,
            path_dem_in,
            mask=mask,
            dem=dem,
        )

    if mode == "using_outlet_pt":
        define_project_extent_using_outlet_point(
            grassdb,
            grass_location,
            qgis_prefix_path,
            path_dem_in,
            outlet_pt,
            buffer_distance=buffer_distance,
            mask="MASK",
            dem="dem",
        )

    if mode == "using_provided_ply":
        define_project_extent_using_input_polygon(
            grassdb,
            grass_location,
            qgis_prefix_path,
            path_dem_in,
            path_extent_ply,
            mask="MASK",
            dem="dem",
        )

    if mode == "using_hybasin":
        define_project_extent_using_hybasin_ply(
            grassdb,
            grass_location,
            qgis_prefix_path,
            path_dem_in,
            buffer_distance=buffer_distance,
            hybasin_ply=hybasin_ply,
            down_hybasin_id=down_hybasin_id,
            up_hybasin_id=up_hybasin_id,
            mask="MASK",
            dem="dem",
        )



import os


def combine_catchments_covered_by_the_same_lake_method(
    Routing_Product_Folder = '#',
    qgis_prefix_path="#",
    area_thresthold = 10*30*30/1000/1000,
    gis_platform="qgis",
):

    if gis_platform == "qgis":
        assert (
            qgis_prefix_path != "#"
        ), "qgis prefix path is needed, when gis_platform = qgis"
        from basinmaker.postprocessing.combine import (
            combine_catchments_covered_by_the_same_lake_qgis,
        )

        combine_catchments_covered_by_the_same_lake_qgis(
            # Path_final_rivply=Path_final_rivply,
            # Path_final_riv=Path_final_riv,
            Routing_Product_Folder = Routing_Product_Folder,
            qgis_prefix_path=qgis_prefix_path,
        )
    if gis_platform == "arcgis":
        from basinmaker.postprocessing.combinearcgis import (
            combine_catchments_covered_by_the_same_lake_arcgis,
        )

        combine_catchments_covered_by_the_same_lake_arcgis(
            Routing_Product_Folder=Routing_Product_Folder,
        )

    if gis_platform == "purepy":
        from basinmaker.postprocessing.combinepurepy import (
            combine_catchments_covered_by_the_same_lake_purepy,
        )

        combine_catchments_covered_by_the_same_lake_purepy(
            Routing_Product_Folder=Routing_Product_Folder,
            area_thresthold = area_thresthold,
        )


def simplify_routing_structure_by_filter_lakes_method(
    Path_final_riv_ply="#",
    Path_final_riv="#",
    Path_Con_Lake_ply="#",
    Path_NonCon_Lake_ply="#",
    Routing_Product_Folder = '#',
    Thres_Area_Conn_Lakes=-1,
    Thres_Area_Non_Conn_Lakes=-1,
    Selected_Lake_List_in=[],
    OutputFolder="#",
    qgis_prefix_path="#",
    gis_platform="qgis",
):

    if gis_platform == "qgis":
        assert (
            qgis_prefix_path != "#"
        ), "qgis prefix path is needed, when gis_platform = qgis"
        from basinmaker.postprocessing.selectlake import (
            simplify_routing_structure_by_filter_lakes_qgis,
        )

        simplify_routing_structure_by_filter_lakes_qgis(
            Routing_Product_Folder = Routing_Product_Folder,
            Thres_Area_Conn_Lakes=Thres_Area_Conn_Lakes,
            Thres_Area_Non_Conn_Lakes=Thres_Area_Non_Conn_Lakes,
            Selected_Lake_List_in=Selected_Lake_List_in,
            OutputFolder=OutputFolder,
            qgis_prefix_path=qgis_prefix_path,
            gis_platform=gis_platform,
        )
    if gis_platform == "arcgis":
        from basinmaker.postprocessing.selectlakearcgis import (
            simplify_routing_structure_by_filter_lakes_arcgis,
        )

        simplify_routing_structure_by_filter_lakes_arcgis(
            Routing_Product_Folder = Routing_Product_Folder,
            Thres_Area_Conn_Lakes=Thres_Area_Conn_Lakes,
            Thres_Area_Non_Conn_Lakes=Thres_Area_Non_Conn_Lakes,
            Selected_Lake_List_in=Selected_Lake_List_in,
            OutputFolder=OutputFolder,
            qgis_prefix_path=qgis_prefix_path,
            gis_platform=gis_platform,
        )

    if gis_platform == "purepy":
        from basinmaker.postprocessing.selectlakepurepy import (
            simplify_routing_structure_by_filter_lakes_purepy,
        )

        simplify_routing_structure_by_filter_lakes_purepy(
            Routing_Product_Folder = Routing_Product_Folder,
            Thres_Area_Conn_Lakes=Thres_Area_Conn_Lakes,
            Thres_Area_Non_Conn_Lakes=Thres_Area_Non_Conn_Lakes,
            Selected_Lake_List_in=Selected_Lake_List_in,
            OutputFolder=OutputFolder,
            qgis_prefix_path=qgis_prefix_path,
            gis_platform=gis_platform,
        )


def simplify_routing_structure_by_drainage_area_method(
    Routing_Product_Folder='#',
    Area_Min=-1,
    OutputFolder="#",
    qgis_prefix_path="#",
    gis_platform="qgis",
):

    if gis_platform == "qgis":
        assert (
            qgis_prefix_path != "#"
        ), "qgis prefix path is needed, when gis_platform = qgis"
        from basinmaker.postprocessing.increaseda import (
            simplify_routing_structure_by_drainage_area_qgis,
        )

        simplify_routing_structure_by_drainage_area_qgis(
            Routing_Product_Folder = Routing_Product_Folder,
            Area_Min=Area_Min,
            OutputFolder=OutputFolder,
            qgis_prefix_path=qgis_prefix_path,
        )

    if gis_platform == "arcgis":
        from basinmaker.postprocessing.increasedaarcgis import (
            simplify_routing_structure_by_drainage_area_arcgis,
        )

        simplify_routing_structure_by_drainage_area_arcgis(
            Routing_Product_Folder = Routing_Product_Folder,
            Area_Min=Area_Min,
            OutputFolder=OutputFolder,
        )

    if gis_platform == "purepy":
        from basinmaker.postprocessing.increasedapurepy import (
            simplify_routing_structure_by_drainage_area_purepy,
        )

        simplify_routing_structure_by_drainage_area_purepy(
            Routing_Product_Folder = Routing_Product_Folder,
            Area_Min=Area_Min,
            OutputFolder=OutputFolder,
        )

def select_part_of_routing_product_method(
    Path_Points,
    Gauge_NMS,
    OutputFolder,
    mostdownid,
    mostupid,
    Path_Catchment_Polygon="#",
    Path_River_Polyline="#",
    Path_Con_Lake_ply="#",
    Path_NonCon_Lake_ply="#",
    qgis_prefix_path="#",
    Routing_Product_Folder = '#',
    gis_platform="qgis",
):

    if gis_platform == "qgis":
        from basinmaker.postprocessing.selectprod import (
            Locate_subid_needsbyuser_qgis,
            Select_Routing_product_based_SubId_qgis,
        )

        if Gauge_NMS[0] != "#" or Path_Points != "#":
            subids = Locate_subid_needsbyuser_qgis(
                Path_Points=Path_Points,
                Gauge_NMS=Gauge_NMS,
                Path_products=Path_Catchment_Polygon,
                qgis_prefix_path=qgis_prefix_path,
            )
            if len(subids) <= 0:
                print("subbasin did not found        ")
                return

            subid = subids[0]
            Select_Routing_product_based_SubId_qgis(
                OutputFolder=OutputFolder,
                Path_Catchment_Polygon=Path_Catchment_Polygon,
                Path_River_Polyline=Path_River_Polyline,
                Path_Con_Lake_ply=Path_Con_Lake_ply,
                Path_NonCon_Lake_ply=Path_NonCon_Lake_ply,
                mostdownid=subid,
                mostupstreamid=-1,
                qgis_prefix_path=qgis_prefix_path,
            )

        else:
            Select_Routing_product_based_SubId_qgis(
                OutputFolder=OutputFolder,
                # Path_Catchment_Polygon=Path_Catchment_Polygon,
                # Path_River_Polyline=Path_River_Polyline,
                # Path_Con_Lake_ply=Path_Con_Lake_ply,
                # Path_NonCon_Lake_ply=Path_NonCon_Lake_ply,
                mostdownid=mostdownid,
                mostupstreamid=mostupid,
                Routing_Product_Folder = Routing_Product_Folder,
                qgis_prefix_path=qgis_prefix_path,
            )

    if gis_platform == "arcgis":
        from basinmaker.postprocessing.selectprodarcgis import (
        Select_Routing_product_based_SubId_arcgis
        )
        Select_Routing_product_based_SubId_arcgis(
            OutputFolder=OutputFolder,
            mostdownid=mostdownid,
            mostupstreamid=mostupid,
            Routing_Product_Folder = Routing_Product_Folder,
        )

    if gis_platform == "purepy":
        from basinmaker.postprocessing.selectprodpurepy import (
        Select_Routing_product_based_SubId_purepy
        )
        Select_Routing_product_based_SubId_purepy(
            OutputFolder=OutputFolder,
            mostdownid=mostdownid,
            mostupstreamid=mostupid,
            Routing_Product_Folder = Routing_Product_Folder,
        )

    return


def generate_hrus_method(
    Path_Subbasin_Ply,
    Landuse_info,
    Soil_info,
    Veg_info,
    Sub_Lake_ID="HyLakeId",
    Sub_ID="SubId",
    Path_Connect_Lake_ply="#",
    Path_Non_Connect_Lake_ply="#",
    Lake_Id="Hylak_id",
    Path_Landuse_Ply="#",
    Landuse_ID="Landuse_ID",
    Path_Soil_Ply="#",
    Soil_ID="Soil_ID",
    Path_Veg_Ply="#",
    Veg_ID="Veg_ID",
    Path_Other_Ply_1="#",
    Other_Ply_ID_1="O_ID_1",
    Path_Other_Ply_2="#",
    Other_Ply_ID_2="O_ID_2",
    DEM="#",
    Inmportance_order = ["Hylak_id","Landuse_ID"],
    min_hru_area_pct_sub = 0.2,
    Project_crs="EPSG:3573",
    OutputFolder="#",
    qgis_prefix_path="#",
    gis_platform="qgis",
    pixel_size = 30,
    area_ratio_thresholds = [0,0,0]
):
    if gis_platform == "qgis":
        from basinmaker.postprocessing.hru import GenerateHRUS_qgis

        GenerateHRUS_qgis(
            Path_Subbasin_Ply=Path_Subbasin_Ply,
            Landuse_info=Landuse_info,
            Soil_info=Soil_info,
            Veg_info=Veg_info,
            Sub_Lake_ID=Sub_Lake_ID,
            Sub_ID=Sub_ID,
            Path_Connect_Lake_ply=Path_Connect_Lake_ply,
            Path_Non_Connect_Lake_ply=Path_Non_Connect_Lake_ply,
            Lake_Id=Lake_Id,
            Path_Landuse_Ply=Path_Landuse_Ply,
            Landuse_ID=Landuse_ID,
            Path_Soil_Ply=Path_Soil_Ply,
            Soil_ID=Soil_ID,
            Path_Veg_Ply=Path_Veg_Ply,
            Veg_ID=Veg_ID,
            Path_Other_Ply_1=Path_Other_Ply_1,
            Other_Ply_ID_1=Other_Ply_ID_1,
            Path_Other_Ply_2=Path_Other_Ply_2,
            Other_Ply_ID_2=Other_Ply_ID_2,
            DEM=DEM,
            Project_crs=Project_crs,
            OutputFolder=OutputFolder,
            qgis_prefix_path=qgis_prefix_path,
            importance_order = Inmportance_order,
            min_hru_pct_sub_area = min_hru_area_pct_sub,
        )

    if gis_platform == "arcgis":
        from basinmaker.postprocessing.hruarcgis import GenerateHRUS_arcgis

        GenerateHRUS_arcgis(
            Path_Subbasin_Ply=Path_Subbasin_Ply,
            Landuse_info=Landuse_info,
            Soil_info=Soil_info,
            Veg_info=Veg_info,
            Sub_Lake_ID=Sub_Lake_ID,
            Sub_ID=Sub_ID,
            Path_Connect_Lake_ply=Path_Connect_Lake_ply,
            Path_Non_Connect_Lake_ply=Path_Non_Connect_Lake_ply,
            Lake_Id=Lake_Id,
            Path_Landuse_Ply=Path_Landuse_Ply,
            Landuse_ID=Landuse_ID,
            Path_Soil_Ply=Path_Soil_Ply,
            Soil_ID=Soil_ID,
            Path_Veg_Ply=Path_Veg_Ply,
            Veg_ID=Veg_ID,
            Path_Other_Ply_1=Path_Other_Ply_1,
            Other_Ply_ID_1=Other_Ply_ID_1,
            Path_Other_Ply_2=Path_Other_Ply_2,
            Other_Ply_ID_2=Other_Ply_ID_2,
            Inmportance_order = Inmportance_order,
            min_hru_area_pct_sub = min_hru_area_pct_sub,
            DEM=DEM,
            Project_crs=Project_crs,
            OutputFolder=OutputFolder,
        )


    if gis_platform == "purepy":
        from basinmaker.postprocessing.hrupurepy import GenerateHRUS_purepy

        GenerateHRUS_purepy(
            Path_Subbasin_Ply=Path_Subbasin_Ply,
            Landuse_info=Landuse_info,
            Soil_info=Soil_info,
            Veg_info=Veg_info,
            Sub_Lake_ID=Sub_Lake_ID,
            Sub_ID=Sub_ID,
            Path_Connect_Lake_ply=Path_Connect_Lake_ply,
            Path_Non_Connect_Lake_ply=Path_Non_Connect_Lake_ply,
            Lake_Id=Lake_Id,
            Path_Landuse_Ply=Path_Landuse_Ply,
            Landuse_ID=Landuse_ID,
            Path_Soil_Ply=Path_Soil_Ply,
            Soil_ID=Soil_ID,
            Path_Veg_Ply=Path_Veg_Ply,
            Veg_ID=Veg_ID,
            Path_Other_Ply_1=Path_Other_Ply_1,
            Other_Ply_ID_1=Other_Ply_ID_1,
            Path_Other_Ply_2=Path_Other_Ply_2,
            Other_Ply_ID_2=Other_Ply_ID_2,
            Inmportance_order = Inmportance_order,
            min_hru_area_pct_sub = min_hru_area_pct_sub,
            DEM=DEM,
            Project_crs=Project_crs,
            OutputFolder=OutputFolder,
            pixel_size = pixel_size,
            area_ratio_thresholds = area_ratio_thresholds,
        )


    return


def obtain_grids_polygon_from_netcdf_file(
    qgis_prefix_path,
    netcdf_path="#",
    output_folder="#",
    coor_x_nm="lon",
    coor_y_nm="lat",
    is_rotated_grid=1,
    r_coor_x_nm="rlon",
    r_coor_y_nm="rlat",
    spatial_ref="EPSG:4326",
    x_add=-360,
    y_add=0,
    gis_platform = "qgis",
):



    if gis_platform == "qgis":
        from basinmaker.postprocessing.gridweight import (
            Generate_Grid_Poly_From_NetCDF_QGIS,
        )
        Generate_Grid_Poly_From_NetCDF_QGIS(
            NetCDF_Path=netcdf_path,
            Output_Folder=output_folder,
            Coor_x_NM=coor_x_nm,
            Coor_y_NM=coor_y_nm,
            Is_Rotated_Grid=is_rotated_grid,
            R_Coor_x_NM=r_coor_x_nm,
            R_Coor_y_NM=r_coor_y_nm,
            SpatialRef=spatial_ref,
            x_add=x_add,
            y_add=y_add,
            qgis_prefix_path=qgis_prefix_path,
        )



def generate_area_weight_of_two_polygons(
    target_polygon_path="#",
    mapping_polygon_path="#",
    col_nm="HRU_ID",
    output_folder="#",
    qgis_prefix_path = '#',
    gis_platform = "qgis"
):

    if gis_platform == "qgis":
        from basinmaker.postprocessing.gridweight import (
            Area_Weighted_Mapping_Between_Two_Polygons_QGIS,
        )
        Area_Weighted_Mapping_Between_Two_Polygons_QGIS(
            Target_Ply_Path=target_polygon_path,
            Mapping_Ply_Path=mapping_polygon_path,
            Col_NM=col_nm,
            Output_Folder=output_folder,
            qgis_prefix_path = qgis_prefix_path,
        )

    if gis_platform == "arcgis":
        from basinmaker.postprocessing.gridweightarcgis import (
            Area_Weighted_Mapping_Between_Two_Polygons_Arcgis,
        )
        Area_Weighted_Mapping_Between_Two_Polygons_Arcgis(
            Target_Ply_Path=target_polygon_path,
            Mapping_Ply_Path=mapping_polygon_path,
            Col_NM=col_nm,
            Output_Folder=output_folder,
        )




import numpy as np
import geopandas
import sys
import os
import csv
import pandas as pd
import tempfile 
import copy 

import sys, os
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

from basinmaker.func.purepy import *
from basinmaker.func.pdtable import *



def Select_Routing_product_based_SubId_purepy(
    OutputFolder,
    Routing_Product_Folder,
    mostdownid=-1,
    mostupstreamid=-1,
):
    """Extract region of interest based on provided Subid

    Function that used to obtain the region of interest from routing
    product based on given SubId

    Parameters
    ----------
    OutputFolder                   : string
        Folder path that stores extracted routing product
    Path_Catchment_Polygon         : string
        Path to the catchment polygon
    Path_River_Polyline            : string (optional)
        Path to the river polyline
    Path_Con_Lake_ply              : string (optional)
        Path to a connected lake polygon. Connected lakes are lakes that
        are connected by Path_final_cat_riv or Path_final_riv.
    Path_NonCon_Lake_ply           : string (optional)
        Path to a non connected lake polygon. Connected lakes are lakes
        that are not connected by Path_final_cat_riv or Path_final_riv.
    mostdownid                     : integer
        It is the most downstream subbasin ID in the region of interest
    mostupstreamid                 : integer (optional)
        It is the most upstream subbasin ID in the region of interest.
        Normally it is -1, indicating all subbasin drainage to mostdownid
        is needed. In some case, if not all subbasin drainage to mostdownid
        is needed, then the most upstream subbsin ID need to be provided
        here.


    Notes
    -------
    This function has no return values, instead following fiels will be
    generated. The output files have are same as inputs expect the extent
    are different.

    os.path.join(OutputFolder,os.path.basename(Path_Catchment_Polygon))
    os.path.join(OutputFolder,os.path.basename(Path_River_Polyline))
    os.path.join(OutputFolder,os.path.basename(Path_Con_Lake_ply))
    os.path.join(OutputFolder,os.path.basename(Path_NonCon_Lake_ply))

    Returns:
    -------
    None

    Examples
    -------

    """
    tempfolder = os.path.join(
        tempfile.gettempdir(),
        "basinmaker_locsubid" + str(101),#np.random.randint(1, 10000 + 1)),
    )
    if not os.path.exists(tempfolder):
        os.makedirs(tempfolder)


    Path_Catchment_Polygon="#"
    Path_River_Polyline="#"
    Path_Con_Lake_ply="#"
    Path_NonCon_Lake_ply="#"
    Path_obs_gauge_point="#"
    Path_final_cat_ply="#"
    Path_final_cat_riv="#"

    ##define input files from routing prodcut 
    for file in os.listdir(Routing_Product_Folder):
        if file.endswith(".shp"):
            if 'catchment_without_merging_lakes' in file:
                Path_Catchment_Polygon = os.path.join(Routing_Product_Folder, file)
            if 'river_without_merging_lakes' in file:
                Path_River_Polyline = os.path.join(Routing_Product_Folder, file)
            if 'sl_connected_lake' in file:
                Path_Con_Lake_ply = os.path.join(Routing_Product_Folder, file)
            if 'sl_non_connected_lake' in file:
                Path_NonCon_Lake_ply = os.path.join(Routing_Product_Folder, file)
            if 'obs_gauges' in file:
                Path_obs_gauge_point = os.path.join(Routing_Product_Folder, file)
            if 'finalcat_info' in file:
                Path_final_cat_ply = os.path.join(Routing_Product_Folder, file)
            if 'finalcat_info_riv' in file:
                Path_final_cat_riv = os.path.join(Routing_Product_Folder, file)                

    if Path_Catchment_Polygon == '#' or  Path_River_Polyline =='#':
        print("Invalid routing product folder ")
        return()



    sub_colnm = "SubId"
    down_colnm = "DowSubId"

    ##3
     
    cat_ply = geopandas.read_file(Path_Catchment_Polygon)

    hyshdinfo = cat_ply[[sub_colnm, down_colnm]].astype("int32").values

    Gauge_col_Name = "Has_POI"
    if "Has_POI" not in cat_ply.columns:
        Gauge_col_Name = "Has_Gauge"
        
    ### Loop for each downstream id

    if not os.path.exists(OutputFolder):
        os.makedirs(OutputFolder)

    ## find all subid control by this subid
    for i_down in range(0,len(mostdownid)):
        ### Loop for each downstream id
        OutHyID = mostdownid[i_down]
        OutHyID2 = mostupstreamid[i_down]
            
        ## find all subid control by this subid
        HydroBasins1 = defcat(hyshdinfo, OutHyID)
        if OutHyID2 > 0:
            HydroBasins2 = defcat(hyshdinfo, OutHyID2)
            ###  exculde the Ids in HydroBasins2 from HydroBasins1
            for i in range(len(HydroBasins2)):
                rows = np.argwhere(HydroBasins1 == HydroBasins2[i])
                HydroBasins1 = np.delete(HydroBasins1, rows)
            HydroBasins = HydroBasins1
        else:
            HydroBasins = HydroBasins1

        if i_down == 0:
            HydroBasins_All = HydroBasins
        else:
            HydroBasins_All = np.concatenate((HydroBasins_All, HydroBasins), axis=0)

    Outputfilename_cat = os.path.join(
        OutputFolder, os.path.basename(Path_Catchment_Polygon)
    )

    cat_ply_select = cat_ply.loc[cat_ply['SubId'].isin(HydroBasins_All)]

    cat_ply_select.to_file(Outputfilename_cat) 
     
    Outputfilename_cat_riv = os.path.join(
        OutputFolder, os.path.basename(Path_River_Polyline)
    )

    cat_riv = geopandas.read_file(Path_River_Polyline)
    

    cat_riv_select = cat_riv.loc[cat_riv['SubId'].isin(HydroBasins)]
    
    cat_riv_select.to_file(Outputfilename_cat_riv) 
    
    cat_ply_select = geopandas.read_file(Outputfilename_cat)
    Connect_Lake_info = cat_ply_select.loc[cat_ply_select["Lake_Cat"] == 1]
    Connect_Lakeids = np.unique(Connect_Lake_info["HyLakeId"].values)
    Connect_Lakeids = Connect_Lakeids[Connect_Lakeids > 0]
    
    NConnect_Lake_info = cat_ply_select.loc[cat_ply_select["Lake_Cat"] == 2]
    NonCL_Lakeids = np.unique(NConnect_Lake_info["HyLakeId"].values)
    NonCL_Lakeids = NonCL_Lakeids[NonCL_Lakeids > 0]

    if len(Connect_Lakeids) > 0 and Path_Con_Lake_ply != "#":

        sl_con_lakes = geopandas.read_file(Path_Con_Lake_ply)
        sl_con_lakes = sl_con_lakes.loc[sl_con_lakes['Hylak_id'].isin(Connect_Lakeids)]
        sl_con_lakes.to_file(os.path.join(OutputFolder,os.path.basename(Path_Con_Lake_ply)))


    if len(NonCL_Lakeids) > 0 and Path_NonCon_Lake_ply != "#":
        sl_non_con_lakes = geopandas.read_file(Path_NonCon_Lake_ply)
        sl_non_con_lakes = sl_non_con_lakes.loc[sl_non_con_lakes['Hylak_id'].isin(NonCL_Lakeids)]
        sl_non_con_lakes.to_file(os.path.join(OutputFolder,os.path.basename(Path_NonCon_Lake_ply)))
    
    sl_gauge_info = cat_ply_select.loc[cat_ply_select[Gauge_col_Name] > 0]
    sl_gauge_nm = np.unique(sl_gauge_info["Obs_NM"].values)
    sl_gauge_nm = sl_gauge_nm[sl_gauge_nm != 'nan']
    if len(sl_gauge_nm) > 0 and Path_obs_gauge_point !='#':
        all_gauge = geopandas.read_file(Path_obs_gauge_point)
        sl_gauge = all_gauge.loc[all_gauge['Obs_NM'].isin(sl_gauge_nm)]
        sl_gauge.to_file(os.path.join(OutputFolder,os.path.basename(Path_obs_gauge_point)))
    
    return 


################################################################


from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import pandas as pd
import numpy as np
import tempfile
from joblib import Parallel, delayed


def combine_catchments_covered_by_the_same_lake_qgis(
    # OutputFolder, 
    # Path_final_rivply="#", 
    # Path_final_riv="#", 
    Routing_Product_Folder = '#',
    qgis_prefix_path="#"
):
    """Define final lake river routing structure

    Generate the final lake river routing structure by merging subbasin
    polygons that are covered by the same lake.
    The input are the catchment polygons and river segements
    before merging for lakes. The input files can be output of
    any of following functions:
    SelectLakes, Select_Routing_product_based_SubId,
    Customize_Routing_Topology,RoutingNetworkTopologyUpdateToolset_riv
    The result is the final catchment polygon that ready to be used for
    hydrological modeling

    Parameters
    ----------
    OutputFolder                   : string
        Folder name that stores generated extracted routing product
    Path_final_riv_ply             : string
        Path to the catchment polygon which is the routing product
        before merging lakes catchments and need to be processed before
        used. It is the input for simplify the routing product based
        on lake area or drianage area.
        routing product and can be directly used.
    Path_final_riv                 : string
        Path to the river polyline which is the routing product
        before merging lakes catchments and need to be processed before
        used. It is the input for simplify the routing product based
        on lake area or drianage area.

    Notes
    -------
    This function has no return values, instead will generate following
    files. They are catchment polygons and river polylines that can be
    used for hydrological modeling.
    os.path.join(OutputFolder,'finalcat_info.shp')
    os.path.join(OutputFolder,'finalcat_info_riv.shp')

    Returns:
    -------
    None

    Examples
    -------

    """

    QgsApplication.setPrefixPath(qgis_prefix_path, True)
    Qgs = QgsApplication([], False)
    Qgs.initQgis()
    from processing.core.Processing import Processing
    from processing.tools import dataobjects
    from qgis import processing

    feedback = QgsProcessingFeedback()
    Processing.initialize()
    QgsApplication.processingRegistry().addProvider(QgsNativeAlgorithms())
    context = dataobjects.createContext()
    context.setInvalidGeometryCheck(QgsFeatureRequest.GeometryNoCheck)


    Path_Catchment_Polygon="#"
    Path_River_Polyline="#"
    Path_Con_Lake_ply="#"
    Path_NonCon_Lake_ply="#"
    Path_obs_gauge_point="#"
    Path_final_cat_ply="#"
    Path_final_cat_riv="#"

    ##define input files from routing prodcut 
    for file in os.listdir(Routing_Product_Folder):
        if file.endswith(".shp"):
            if 'catchment_without_merging_lakes' in file:
                Path_Catchment_Polygon = os.path.join(Routing_Product_Folder, file)
            if 'river_without_merging_lakes' in file:
                Path_River_Polyline = os.path.join(Routing_Product_Folder, file)
            if 'sl_connected_lake' in file:
                Path_Con_Lake_ply = os.path.join(Routing_Product_Folder, file)
            if 'sl_non_connected_lake' in file:
                Path_NonCon_Lake_ply = os.path.join(Routing_Product_Folder, file)
            if 'obs_gauges' in file:
                Path_obs_gauge_point = os.path.join(Routing_Product_Folder, file)
            if 'finalcat_info' in file:
                Path_final_cat_ply = os.path.join(Routing_Product_Folder, file)
            if 'finalcat_info_riv' in file:
                Path_final_cat_riv = os.path.join(Routing_Product_Folder, file)                

    if Path_Catchment_Polygon == '#' or  Path_River_Polyline =='#':
        print("Invalid routing product folder ")


    OutputFolder = Routing_Product_Folder

    sub_colnm = "SubId"
    Path_final_rviply = Path_Catchment_Polygon
    Path_final_riv = Path_River_Polyline

    if not os.path.exists(OutputFolder):
        os.makedirs(OutputFolder)
    tempfolder = os.path.join(
        tempfile.gettempdir(), "basinmaker_" + str(np.random.randint(1, 10000 + 1))
    )
    if not os.path.exists(tempfolder):
        os.makedirs(tempfolder)

    ### create a copy of shapfiles in temp folder
    Path_Temp_final_rviply = os.path.join(
        tempfolder,
        "temp_finalriv_ply" + str(np.random.randint(1, 10000 + 1)) + ".shp",
    )
    Path_Temp_final_rvi = os.path.join(
        tempfolder,
        "temp_finalriv" + str(np.random.randint(1, 10000 + 1)) + ".shp",
    )
    qgis_vector_dissolve(
        processing,
        context,
        INPUT=Path_final_rviply,
        FIELD=["SubId"],
        OUTPUT=Path_Temp_final_rviply,
    )
    qgis_vector_dissolve(
        processing,
        context,
        INPUT=Path_final_riv,
        FIELD=["SubId"],
        OUTPUT=Path_Temp_final_rvi,
    )

    ### read riv ply info
    ### read attribute table
    finalrivply_info = Dbf_To_Dataframe(Path_Temp_final_rviply).drop_duplicates(
        "SubId", keep="first"
    )
    # change attribute table for lake covered catchments,
    mapoldnew_info = change_attribute_values_for_catchments_covered_by_same_lake(
        finalrivply_info
    )
    
    mapoldnew_info.loc[mapoldnew_info['Lake_Cat'] > 0,'RivLength'] = -1.2345
    mapoldnew_info.loc[mapoldnew_info['Lake_Cat'] > 0,'RivSlope'] = -1.2345
    mapoldnew_info.loc[mapoldnew_info['Lake_Cat'] > 0,'FloodP_n'] = -1.2345
    mapoldnew_info.loc[mapoldnew_info['Lake_Cat'] > 0,'Ch_n'] = -1.2345
    mapoldnew_info.loc[mapoldnew_info['Lake_Cat'] > 0,'Max_DEM'] = -1.2345
    mapoldnew_info.loc[mapoldnew_info['Lake_Cat'] > 0,'Min_DEM'] = -1.2345
    
    if 'DA_Chn_L' in mapoldnew_info.columns:
        mapoldnew_info.loc[mapoldnew_info['Lake_Cat'] > 0,'DA_Chn_L'] = -1.2345
        mapoldnew_info.loc[mapoldnew_info['Lake_Cat'] > 0,'DA_Chn_Slp'] = -1.2345

    # update topology for new attribute table
    update_topology(mapoldnew_info, UpdateStreamorder=-1)

    # copy new attribute table to shpfile
    all_subids = finalrivply_info['SubId'].values
    
    copy_data_and_dissolve(all_subids,tempfolder,processing,Path_Temp_final_rviply,Path_Temp_final_rvi,
        mapoldnew_info,COLUMN_NAMES_CONSTANT_CLEAN,OutputFolder,Path_Catchment_Polygon,context)
        
    return 


from arcgis import GIS
from arcgis.features import GeoAccessor, GeoSeriesAccessor
import arcpy
from arcpy import env
from arcpy.sa import *
import numpy as np
import os
import pandas as pd 
import tempfile
import arcpy.cartography as CA
from json import load, JSONEncoder
import json
import shutil
#####
arcpy.env.overwriteOutput = True
arcpy.CheckOutExtension("Spatial")

def select_feature_by_attributes_arcgis(input,Attri_NM,Attri_v,output):
    where_clause = '"%s" IN' % (Attri_NM) 
    where_clause = where_clause + " (" 
    for i in range(0,len(Attri_v)):
        if i == 0:
            where_clause = where_clause + str(Attri_v[i])
        else:
            where_clause = where_clause + "," + str(Attri_v[i])
    where_clause = where_clause + ")"
    
    arcpy.Select_analysis(input, output, where_clause)
    return
##################


def Remove_Unselected_Lake_Attribute_In_Finalcatinfo_Arcgis(finalcat_ply, Conn_Lake_Ids):
    """Functions will set lake id not in Conn_Lake_Ids to -1.2345 in attribute
        table of Path_Finalcatinfo
    ----------

    Notes
    -------

    Returns:
    -------
        None, the attribute table of Path_shpfile will be updated
    """
    
    mask1 = np.logical_not(finalcat_ply['HyLakeId'].isin(Conn_Lake_Ids))
    mask2 = finalcat_ply['Lake_Cat'] != 2
    mask = np.logical_and(mask1,mask2)
    
    finalcat_ply.loc[mask,'HyLakeId'] = 0
    finalcat_ply.loc[mask,'LakeVol'] = 0
    finalcat_ply.loc[mask,'LakeArea'] = 0
    finalcat_ply.loc[mask,'LakeDepth'] = 0
    finalcat_ply.loc[mask,'Laketype'] =0
    finalcat_ply.loc[mask,'Lake_Cat'] = 0
    
    return finalcat_ply


def save_modified_attributes_to_outputs(mapoldnew_info,tempfolder,OutputFolder,cat_name,riv_name,Path_final_riv,dis_col_name='SubId'):

    mapoldnew_info.spatial.to_featureclass(location=os.path.join(tempfolder,'updateattri.shp'),overwrite=True,sanitize_columns=False)
    arcpy.Dissolve_management(os.path.join(tempfolder,'updateattri.shp'), os.path.join(OutputFolder,cat_name), [dis_col_name])
    arcpy.JoinField_management(os.path.join(OutputFolder,cat_name), dis_col_name, os.path.join(tempfolder,'updateattri.shp'), dis_col_name)
    arcpy.DeleteField_management(os.path.join(OutputFolder,cat_name), 
        ["SubId_1", "Id","nsubid2", "nsubid","ndownsubid","Old_SubId","Old_DowSub","Join_Count","TARGET_FID","Id","SubID_Oldr","HRU_ID_N_1","HRU_ID_N_2","facters"]
    )
    
    
    if riv_name != '#':
        arcpy.CalculateGeometryAttributes_management(os.path.join(OutputFolder, cat_name), [["centroid_x", "CENTROID_X"], ["centroid_y", "CENTROID_Y"]],coordinate_system = arcpy.SpatialReference(4326))
        cat_colnms = mapoldnew_info.columns
        drop_cat_colnms = cat_colnms[cat_colnms.isin(["SHAPE","SubId_1", "Id","nsubid2", "nsubid","ndownsubid","Old_DowSub","Join_Count","TARGET_FID","Id","SubID_Oldr","HRU_ID_N_1","HRU_ID_N_2","facters","Old_DowSubId"])]
        cat_pd = mapoldnew_info.drop(columns=drop_cat_colnms)

        riv_pd = pd.DataFrame.spatial.from_featureclass(Path_final_riv)
        riv_pd['Old_SubId'] = riv_pd['SubId']
        # remove all columns 
        riv_pd = riv_pd[['SHAPE','Old_SubId']]        
        riv_pd = pd.merge(riv_pd, cat_pd, on='Old_SubId', how='left')
        
        riv_pd = riv_pd.drop(columns=['Old_SubId'])
        mask = np.logical_or(riv_pd['RivLength'] > 0,riv_pd['Lake_Cat'] > 0)
        riv_pd = riv_pd[mask]
        riv_pd.drop(columns=['centroid_x','centroid_y'])
        riv_pd.spatial.to_featureclass(location=os.path.join(tempfolder,'riv_attri.shp'),overwrite=True,sanitize_columns=False)
        arcpy.Dissolve_management(os.path.join(tempfolder,'riv_attri.shp'), os.path.join(OutputFolder,riv_name), ["SubId"])
        arcpy.JoinField_management(os.path.join(OutputFolder,riv_name), "SubId", os.path.join(tempfolder,'riv_attri.shp'), "SubId")
        arcpy.DeleteField_management(os.path.join(OutputFolder,riv_name), 
            ["SubId_1", "Id","nsubid2", "nsubid","ndownsubid","Old_SubId","Old_DowSub","Join_Count","TARGET_FID","Id","SubID_Oldr","HRU_ID_N_1","HRU_ID_N_2","facters",'centroid_x','centroid_y']
        )
    
    # if "finalcat_info" in cat_name:
    #     create_geo_jason_file(os.path.join(OutputFolder,cat_name))
    
def clean_attribute_name_arcgis(table,names):
    remove_column_names = table.columns[np.logical_not(np.isin(table.columns,names))]
    table = table.drop(columns=remove_column_names)
    return table 


####
def create_geo_jason_file(Input_Polygon_path):

    arcpy.env.overwriteOutput = True
    product_dir = os.path.dirname(Input_Polygon_path)
    Names_in = os.path.basename(Input_Polygon_path).split('_')
    n_charc = len(Names_in)
    version  = Names_in[n_charc - 1][0:4]
    TOLERANCEs = [0.0001,0.0005,0.001,0.005,0.01,0.05]


    head_name_cat = "finalcat_info"
    head_name_riv = "finalcat_info_riv"
    head_name_slake = "sl_connected_lake"
    head_name_nlake = "sl_non_connected_lake"

    Input_file_name = []
    Output_file_name = []                            
    if 'v' in version:
        Input_file_name = [
                           head_name_cat + "_"+version+'.shp',
                           head_name_riv + "_"+version+'.shp',
                           head_name_slake + "_"+version+'.shp',
                           head_name_nlake + "_"+version+'.shp',
                           ]
        Output_file_name = [
                           head_name_cat + "_"+version+'.geojson',
                           head_name_riv + "_"+version+'.geojson',
                           head_name_slake + "_"+version+'.geojson',
                           head_name_nlake + "_"+version+'.geojson',
                           ]
    else:
        Input_file_name = [
                           head_name_cat +'.shp',
                           head_name_riv +'.shp',
                           head_name_slake +'.shp',
                           head_name_nlake +'.shp',
                           ]
        Output_file_name = [
                           head_name_cat +'.geojson',
                           head_name_riv +'.geojson',
                           head_name_slake +'.geojson',
                           head_name_nlake +'.geojson',
                           ]
    created_jason_files = []   
    created_jason_files_lake_riv = [] 
    
    for i  in  range(0,len(Input_file_name)):
        input_path = os.path.join(product_dir,Input_file_name[i])
        output_jason_path = os.path.join(product_dir,Output_file_name[i])
        if not os.path.exists(input_path):
            continue 
        created_jason_files.append(output_jason_path) 
        
        if 'finalcat_info_riv' in Input_file_name[i] or 'connected_lake' in Input_file_name[i]:
            created_jason_files_lake_riv.append(output_jason_path)  
                            
        # reproject to WGS84
        input_wgs_84 = os.path.join(tempfile.gettempdir(),"input_wgs_84.shp")
        arcpy.Project_management(input_path, input_wgs_84, arcpy.SpatialReference(int(4326)))


        if 'finalcat_info' in Input_file_name[i] or "finalcat_info_riv" in Input_file_name[i]:
            arcpy.AddField_management(input_wgs_84, 'rvhName', "TEXT")
            arcpy.CalculateField_management(input_wgs_84, 'rvhName', "'sub' + str(int(\"!SubId!\"))", "PYTHON3")

        arcpy.RepairGeometry_management(input_wgs_84)

        for TOLERANCE in TOLERANCEs:
            input_wgs_84_simplify = os.path.join(tempfile.gettempdir(),"input_wgs_84_simplify.shp")
            if arcpy.Exists(input_wgs_84_simplify):
                arcpy.Delete_management(input_wgs_84_simplify)
                
                
            if "finalcat_info_riv" not in Input_file_name[i]:
                CA.SimplifyPolygon(input_wgs_84, input_wgs_84_simplify, "POINT_REMOVE", TOLERANCE)
            else:
                CA.SimplifyLine(input_wgs_84, input_wgs_84_simplify, "POINT_REMOVE", TOLERANCE)

            arcpy.FeaturesToJSON_conversion(input_wgs_84_simplify, 
                                            output_jason_path,"FORMATTED","NO_Z_VALUES","NO_M_VALUES","GEOJSON")
                                
            json_file_size = os.stat(output_jason_path).st_size/1024/1024 #to MB
            if json_file_size <= 100:
                break                                

    if len(created_jason_files_lake_riv) > 1 and os.stat(os.path.join(product_dir,Output_file_name[1])).st_size/1024/1024 < 500:
        for i in range(0,len(created_jason_files_lake_riv)):
            injson2 = load(open(created_jason_files_lake_riv[i]))
            if 'finalcat_info_riv' in created_jason_files_lake_riv[i]:
                new_features = []
                for element in injson2["features"]:
                    if element["properties"]["Lake_Cat"] == 0:    
                        new_features.append(element) 
                injson2["features"] = new_features

            if i == 0:
                output_jason_lake_riv = injson2
            else:
                output_jason_lake_riv['features'] += injson2['features']
                
        with open(os.path.join(product_dir,'routing_product_lake_river.geojson'), 'w', encoding='utf-8') as f:
            json.dump(output_jason_lake_riv, f, ensure_ascii=False, indent=4)
    else:
        shutil.copy(created_jason_files_lake_riv[0], os.path.join(product_dir,'routing_product_lake_river.geojson'))             


                                 
    return 


    
    



import sys, os
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
from postprocessing.hruarcgis import GenerateHRUS_arcgis


Path_Subbasin_Ply = sys.argv[1]
Path_Connect_Lake_ply = sys.argv[2]
Path_Non_Connect_Lake_ply = sys.argv[3]
Path_Landuse_Ply = sys.argv[4]
Path_Soil_Ply = sys.argv[5]
Path_Veg_Ply = sys.argv[6]
Path_Other_Ply_1 = sys.argv[7]
Path_Other_Ply_2 = sys.argv[8]
Landuse_info = sys.argv[9]
Soil_info = sys.argv[10]
Veg_info = sys.argv[11]
DEM = sys.argv[12]
Inmportance_order = sys.argv[13]
min_hru_area_pct_sub = float(sys.argv[14])
Project_crs = sys.argv[15]
OutputFolder = sys.argv[16]

Other_Ply_ID_1="O_ID_1"
Other_Ply_ID_2="O_ID_2"
Landuse_ID = 'Landuse_ID'
Soil_ID = 'Soil_ID'
Veg_ID = 'Veg_ID'

Sub_Lake_ID="HyLakeId"
Sub_ID="SubId"
Lake_Id="Hylak_id"
Landuse_ID="Landuse_ID"
Soil_ID="Soil_ID"
Other_Ply_ID_1="O_ID_1"
Veg_ID="Veg_ID"
Other_Ply_ID_2="O_ID_2"
            
Inmportance_order = Inmportance_order.split(";")

Sub_Lake_ID = 'HyLakeId',
Sub_ID = 'SubId',
Lake_Id = 'Hylak_id',

GenerateHRUS_arcgis(
    Path_Subbasin_Ply=Path_Subbasin_Ply,
    Landuse_info=Landuse_info,
    Soil_info=Soil_info,
    Veg_info=Veg_info,
    Sub_Lake_ID=Sub_Lake_ID,
    Sub_ID=Sub_ID,
    Path_Connect_Lake_ply=Path_Connect_Lake_ply,
    Path_Non_Connect_Lake_ply=Path_Non_Connect_Lake_ply,
    Lake_Id=Lake_Id,
    Path_Landuse_Ply=Path_Landuse_Ply,
    Landuse_ID=Landuse_ID,
    Path_Soil_Ply=Path_Soil_Ply,
    Soil_ID=Soil_ID,
    Path_Veg_Ply=Path_Veg_Ply,
    Veg_ID=Veg_ID,
    Path_Other_Ply_1=Path_Other_Ply_1,
    Other_Ply_ID_1=Other_Ply_ID_1,
    Path_Other_Ply_2=Path_Other_Ply_2,
    Other_Ply_ID_2=Other_Ply_ID_2,
    Inmportance_order = Inmportance_order,
    min_hru_area_pct_sub = min_hru_area_pct_sub,
    DEM=DEM,
    Project_crs=Project_crs,
    OutputFolder=OutputFolder,
)


from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import sqlite3


def select_lakes_by_area_r(
    grass,
    con,
    garray,
    str_r,
    threshold_con_lake,
    threshold_non_con_lake,
    lake_attributes,
    lakes="alllake",
    connected_lake="connect_lake",
    non_connected_lake="nonconnect_lake",
    str_connected_lake="str_connected_lake",
    sl_connected_lake="sl_connect_lake",
    sl_non_connected_lake="sl_nonconnect_lake",
    sl_lakes="selected_lakes",
    sl_str_connected_lake="str_sl_connected_lake",
    only_included_lake_at_river_interction=False,
):
    # get lake attributes
    sqlstat = "SELECT %s,%s FROM %s" % (
        lake_attributes[0],
        lake_attributes[2],
        lakes,
    )

    lake_info = pd.read_sql_query(sqlstat, con)
    lake_info = lake_info.fillna(-9999)
    un_sl_con_lakeids = lake_info.loc[
        lake_info[lake_attributes[2]] < threshold_con_lake
    ][lake_attributes[0]].values
    un_sl_noncon_lakeids = lake_info.loc[
        lake_info[lake_attributes[2]] < threshold_non_con_lake
    ][lake_attributes[0]].values
    all_lake_ids = lake_info[lake_attributes[0]].values

    all_connected_lake_array = garray.array(mapname=connected_lake)
    all_non_connected_lake_array = garray.array(mapname=non_connected_lake)

    # remove lakes eixst in raster but not in polygon
    cl_lake_ids = np.unique(all_connected_lake_array)
    cl_lake_ids = cl_lake_ids[cl_lake_ids > 0]
    cl_lake_in_polygon = np.isin(cl_lake_ids, all_lake_ids)
    cl_lake_ids_not_in_ply = cl_lake_ids[np.logical_not(cl_lake_in_polygon)]
    if len(cl_lake_ids_not_in_ply) > 0:
        mask = np.isin(all_connected_lake_array, cl_lake_ids_not_in_ply)
        all_connected_lake_array[mask] = -9999

    # change unselected connected lake to -999
    if len(un_sl_con_lakeids) > 0:
        mask = np.isin(all_connected_lake_array, un_sl_con_lakeids)
        all_connected_lake_array[mask] = -9999

    # remove lakes eixst in raster but not in polygon
    ncl_lake_ids = np.unique(all_non_connected_lake_array)
    ncl_lake_ids = ncl_lake_ids[ncl_lake_ids > 0]
    ncl_lake_in_polygon = np.isin(ncl_lake_ids, all_lake_ids)
    ncl_lake_ids_not_in_ply = ncl_lake_ids[np.logical_not(ncl_lake_in_polygon)]
    if len(ncl_lake_ids_not_in_ply) > 0:
        mask = np.isin(all_non_connected_lake_array, ncl_lake_ids_not_in_ply)
        all_non_connected_lake_array[mask] = -9999

    # change unselected connected lake to -9999
    if len(un_sl_noncon_lakeids) > 0:
        mask = np.isin(all_non_connected_lake_array, un_sl_noncon_lakeids)
        all_non_connected_lake_array[mask] = -9999

    temparray = garray.array()

    temparray[:, :] = all_connected_lake_array[:, :]
    temparray.write(mapname=sl_connected_lake, overwrite=True)
    grass.run_command("r.null", map=sl_connected_lake, setnull=[-9999, 0])
    exp = "%s = int(%s)" % (
        sl_connected_lake,
        sl_connected_lake,
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    temparray[:, :] = all_non_connected_lake_array[:, :]
    temparray.write(mapname=sl_non_connected_lake, overwrite=True)
    grass.run_command("r.null", map=sl_non_connected_lake, setnull=[-9999, 0])
    exp = "%s = int(%s)" % (
        sl_non_connected_lake,
        sl_non_connected_lake,
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    exp = "%s = if(isnull(%s),int(%s),int(%s))" % (
        sl_lakes,
        sl_connected_lake,
        sl_non_connected_lake,
        sl_connected_lake,
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)
        
    if only_included_lake_at_river_interction:
        # first find lakes id and river reach id 
        
        lakeid, str_id = generate_stats_list_from_grass_raster(
            grass, mode=2, input_a=sl_lakes, input_b=str_r
        )
        lakeid = np.array(lakeid)
        (unique, counts) = np.unique(lakeid, return_counts=True)
        lake_counts = np.column_stack((unique, counts))
        # find lake id only occurs <=1
        Remove_lake_ids = lake_counts[lake_counts[:,1] <= 1][:,0]
        if len(Remove_lake_ids) > 0:
            grass.run_command(
                "r.null", map=sl_connected_lake, setnull=Remove_lake_ids, overwrite=True
            )

            grass.run_command(
                "r.null", map=sl_lakes, setnull=Remove_lake_ids, overwrite=True
            )
            
    exp = "%s = if(isnull(int(%s)),null(),%s)" % (
        sl_str_connected_lake,
        str_r,
        sl_connected_lake,
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

#        print(unique)
#        print(Remove_lake_ids)
       #if a lake cover more than two times 
       # means this lake cover more than two river 
            
        
    return



from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import sqlite3
from scipy.optimize import curve_fit
from basinmaker.preprocessing.reprojectandclipvectorbyplyqgis import (
    reproject_clip_vectors_by_polygon,
)


def calculate_bankfull_width_depth_from_polyline(
    grassdb,
    grass_location,
    qgis_prefix_path,
    path_bkfwidthdepth,
    path_k_c_zone_polygon,
    bkfwd_attributes,
    catinfo,
    input_geo_names,
    k_in=-1,
    c_in=-1,
    return_k_c_only=False,
):
    mask = input_geo_names["mask"]
    cat_ply_info = input_geo_names["cat_ply_info"]

    default_slope = min_riv_slope
    min_manning_n = 0.025
    max_manning_n = 0.15
    default_bkf_width = 1.2345
    default_bkf_depth = 1.2345
    default_bkf_q = 1.2345
    k = -1
    c = -1
    if path_bkfwidthdepth != "#":
        reproject_clip_vectors_by_polygon(
            grassdb=grassdb,
            grass_location=grass_location,
            qgis_prefix_path=qgis_prefix_path,
            mask=os.path.join(grassdb, mask + ".shp"),
            path_polygon=path_bkfwidthdepth,
            ply_name="bkf_width_depth",
        )
    if path_k_c_zone_polygon != '#':
        reproject_clip_vectors_by_polygon(
            grassdb=grassdb,
            grass_location=grass_location,
            qgis_prefix_path=qgis_prefix_path,
            mask=os.path.join(grassdb, mask + ".shp"),
            path_polygon=path_k_c_zone_polygon,
            ply_name="kc_zone",
        )

    import grass.script as grass
    import grass.script.setup as gsetup
    from grass.pygrass.modules import Module
    from grass.pygrass.modules.shortcuts import general as g
    from grass.pygrass.modules.shortcuts import raster as r
    from grass.script import array as garray
    from grass.script import core as gcore
    from grass_session import Session

    os.environ.update(
        dict(GRASS_COMPRESS_NULLS="1", GRASS_COMPRESSOR="ZSTD", GRASS_VERBOSE="1")
    )
    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")

    con = sqlite3.connect(
        os.path.join(grassdb, grass_location, "PERMANENT", "sqlite", "sqlite.db")
    )

    if path_bkfwidthdepth != "#":
        grass.run_command(
            "v.import",
            input=os.path.join(grassdb, "bkf_width_depth" + ".shp"),
            output="bk_full_wid_depth",
            overwrite=True,
        )

    if path_k_c_zone_polygon != "#":
        grass.run_command(
            "v.import",
            input=os.path.join(grassdb, "kc_zone_clip" + ".shp"),
            output="kc_zone",
            overwrite=True,
        )

        grass.run_command(
            "v.overlay",
            ainput="kc_zone",
            alayer=1,
            atype="area",
            binput=cat_ply_info,
            operator="and",
            output=cat_ply_info + "kc",
            overwrite=True,
        )
        grass.run_command(
            "v.to.db",
            map=cat_ply_info + "kc",
            option="area",
            columns="Area_kc",
            units="meters",
            overwrite=True,
        )


    if k_in < 0 and c_in < 0:
        ### read catchment
        if path_bkfwidthdepth != '#':
            sqlstat = "SELECT %s,%s,%s,%s FROM %s" % (
                bkfwd_attributes[3],
                bkfwd_attributes[2],
                bkfwd_attributes[1],
                bkfwd_attributes[0],
                "bk_full_wid_depth",
            )
            bkf_width_depth = pd.read_sql_query(sqlstat, con)
            bkf_width_depth = bkf_width_depth.fillna(-9999)

            da_q = bkf_width_depth[[bkfwd_attributes[3], bkfwd_attributes[2]]].values

            if len(da_q) > 3:
                k, c = return_k_and_c_in_q_da_relationship(da_q)
            elif len(da_q) > 0 and len(da_q) <= 3:
                k = -1
                c = -1
                default_bkf_width = np.average(bkf_width_depth[bkfwd_attributes[0]])
                default_bkf_depth = np.average(bkf_width_depth[bkfwd_attributes[1]])
                default_bkf_q = np.average(bkf_width_depth[bkfwd_attributes[3]])
            else:
                k = -1
                c = -1

        if path_k_c_zone_polygon != '#':
            k = -1
            c = -1
            sqlstat = "SELECT a_k, a_c,b_Gridcode, Area_kc FROM %s" % (cat_ply_info + "kc")
            k_c_sub = pd.read_sql_query(sqlstat, con)
            k_c_sub = k_c_sub.fillna(-9999)
            k_c_sub = k_c_sub.sort_values(by='Area_kc', ascending=False)
            k_c_sub = k_c_sub.drop_duplicates(subset=['b_Gridcode'])

    else:
        k = k_in
        c = c_in

    if return_k_c_only:
        return k, c


    if 'k_c_sub' not in locals() and k == -1:
        return

    idx = catinfo.index
    for i in range(0, len(idx)):
        idx_i = idx[i]
        da = catinfo.loc[idx_i, "DrainArea"] / 1000 / 1000  # m2 to km2
        catid = catinfo.loc[idx_i, "SubId"]
        if k > 0:
            q = func_Q_DA(da, k, c)
            catinfo.loc[idx_i, "BkfWidth"] =  max(7.2 * q ** 0.5,min_bkf_width)
            catinfo.loc[idx_i, "BkfDepth"] =  max(0.27 * q ** 0.3,min_bkf_depth)
            catinfo.loc[idx_i, "Q_Mean"] = q
        elif 'k_c_sub' in locals():
            if len(k_c_sub.loc[k_c_sub["b_Gridcode"] == catid]) < 1:
                k_sub =  0.00450718
                c_sub =  0.98579699
                print("k_sub not found .....")
            else:
                k_sub = k_c_sub.loc[k_c_sub["b_Gridcode"] == catid]["a_k"].values[0]
                c_sub = k_c_sub.loc[k_c_sub["b_Gridcode"] == catid]["a_c"].values[0]
                if k_sub < 0:
                    k_sub =  0.00450718
                    c_sub =  0.98579699

            q = func_Q_DA(da, k_sub, c_sub)
            catinfo.loc[idx_i, "BkfWidth"] =  max(7.2 * q ** 0.5,min_bkf_width)
            catinfo.loc[idx_i, "BkfDepth"] =  max(0.27 * q ** 0.3,min_bkf_depth)
            catinfo.loc[idx_i, "Q_Mean"] = q

        else:
            catinfo.loc[idx_i, "BkfWidth"] = default_bkf_width
            catinfo.loc[idx_i, "BkfDepth"] = default_bkf_depth
            catinfo.loc[idx_i, "Q_Mean"] = default_bkf_q

        if catinfo.loc[idx_i, "Lake_Cat"] < 2:
            if catinfo.loc[idx_i, "Max_DEM"] < 0:
                catinfo.loc[idx_i, "Max_DEM"] = catinfo.loc[idx_i, "MeanElev"]
                catinfo.loc[idx_i, "Min_DEM"] = catinfo.loc[idx_i, "MeanElev"]

    # adjust channel parameters

    #remove ncl and headwater subbasins
    catinfo_riv = catinfo.loc[(catinfo["Lake_Cat"] < 2) & (catinfo["RivLength"] != -1.2345)].copy(deep=True)

    Seg_IDS = catinfo_riv["Seg_ID"].values
    Seg_IDS = np.unique(Seg_IDS)

    for iseg in range(0, len(Seg_IDS)):
        i_seg_id = Seg_IDS[iseg]
        i_seg_info = catinfo_riv[catinfo_riv["Seg_ID"] == i_seg_id]
        if len(i_seg_info["RivLength"].values[i_seg_info["RivLength"].values > 0]) > 0:
            length_seg = np.sum(i_seg_info["RivLength"].values[i_seg_info["RivLength"].values > 0])
        else:
            length_seg = 1
        qmean_seg = np.average(i_seg_info["Q_Mean"].values[i_seg_info["Q_Mean"].values > 0])
        width_seg = np.average(i_seg_info["BkfWidth"].values[i_seg_info["BkfWidth"].values > 0])
        depth_Seg = np.average(i_seg_info["BkfDepth"].values[i_seg_info["BkfDepth"].values > 0])
        floodn_Seg = np.average(i_seg_info["FloodP_n"].values[i_seg_info["FloodP_n"].values > 0])
        basslp_Seg = np.average(i_seg_info["BasSlope"].values[i_seg_info["BasSlope"].values > 0])
        basasp_Seg = np.average(i_seg_info["BasAspect"].values[i_seg_info["BasAspect"].values > 0])
        baselv_Seg = np.average(i_seg_info["MeanElev"].values[i_seg_info["MeanElev"].values > 0])

        if len(i_seg_info["Max_DEM"].values[i_seg_info["Max_DEM"].values > -999999999999999999]) > 0:
            max_elve_seg = np.max(i_seg_info["Max_DEM"].values[i_seg_info["Max_DEM"].values > -9999999999999999])
            min_elve_seg = np.min(i_seg_info["Min_DEM"].values[i_seg_info["Min_DEM"].values > -9999999999999999])
        else:
            max_elve_seg = baselv_Seg
            min_elve_seg = baselv_Seg

        slope_seg = (max_elve_seg - min_elve_seg) / length_seg
        if slope_seg < 0.000000001:
            slope_seg = min_riv_slope  #### Needs to update later

        n_seg = calculateChannaln(width_seg, depth_Seg, qmean_seg, slope_seg)

        for i in range(0, len(i_seg_info)):
            subid = i_seg_info["SubId"].values[i]
            max_elve_rch = i_seg_info["Max_DEM"].values[i]
            min_elve_rch = i_seg_info["Min_DEM"].values[i]
            length_rch = max(i_seg_info["RivLength"].values[i],min_riv_lenth)
            qmean_rch = i_seg_info["Q_Mean"].values[i]
            width_rch = i_seg_info["BkfWidth"].values[i]
            depth_rch = i_seg_info["BkfDepth"].values[i]
            floodn_rch = i_seg_info["FloodP_n"].values[i]

            if min_elve_rch < -2000:
                if i_seg_info["MeanElev"].values[i] > -2000:
                    min_elve_rch = i_seg_info["MeanElev"].values[i]
                else:
                    min_elve_rch = baselv_Seg

            if max_elve_rch < -2000:
                if i_seg_info["MeanElev"].values[i] > -2000:
                    max_elve_rch = i_seg_info["MeanElev"].values[i]
                else:
                    max_elve_rch = baselv_Seg

            slope_rch = (max_elve_rch - min_elve_rch) / length_rch


            if slope_rch < min_riv_slope or slope_rch > max_riv_slope:
                if slope_seg >= min_riv_slope and slope_seg <= max_riv_slope:
                    slope_rch = slope_seg

            slope_rch = max(slope_rch,min_riv_slope)
            slope_rch = min(slope_rch,max_riv_slope)

            n_rch = calculateChannaln(width_rch, depth_rch, qmean_rch, slope_rch)

            if n_rch < min_manning_n or n_rch > max_manning_n:
                if n_seg >= min_manning_n and n_seg <= max_manning_n:
                    n_rch = n_seg

            n_rch = max(n_rch,min_manning_n)
            n_rch = min(n_rch,max_manning_n)

            catinfo.loc[catinfo["SubId"] == subid, "RivSlope"] = slope_rch
            catinfo.loc[catinfo["SubId"] == subid, "Ch_n"] = n_rch
            catinfo.loc[catinfo["SubId"] == subid, "RivLength"] = length_rch


            if subid == 504:
                print(floodn_rch)


            if floodn_rch < 0:
                if floodn_Seg > 0:
                    floodn_rch = floodn_Seg
                else:
                    floodn_rch = DEFALUT_FLOOD_N

            floodn_rch = max(floodn_rch,n_rch)
            catinfo.loc[catinfo["SubId"] == subid, "FloodP_n"] = floodn_rch
            catinfo.loc[catinfo["SubId"] == subid, "Max_DEM"] = max_elve_rch
            catinfo.loc[catinfo["SubId"] == subid, "Min_DEM"] = min_elve_rch

            # if subid == 504:
            #     print(floodn_rch)
            #     asdf

            if i_seg_info["BasSlope"].values[i] <= 0:
                if basslp_Seg > 0:
                    catinfo.loc[catinfo["SubId"] == subid, "BasSlope"] = basslp_Seg
                else:
                    catinfo.loc[catinfo["SubId"] == subid, "BasSlope"] = 0

            if i_seg_info["BasAspect"].values[i] <= 0:
                if basasp_Seg > 0:
                    catinfo.loc[catinfo["SubId"] == subid, "BasAspect"] = basasp_Seg
                else:
                    catinfo.loc[catinfo["SubId"] == subid, "BasAspect"] = 0

            if i_seg_info["MeanElev"].values[i] < 0:
                if baselv_Seg > 0:
                    catinfo.loc[catinfo["SubId"] == subid, "MeanElev"] = baselv_Seg
                else:
                    catelv = catinfo['MeanElev'].values
                    catelv = catelv[catelv > 0]
                    catinfo.loc[catinfo["SubId"] == subid, "MeanElev"] =np.average(catelv)
    PERMANENT.close()
    return catinfo
